"""
.. currentmodule:: neet.boolean.wtnetwork

.. testsetup:: wtnetwork

    from neet.boolean.examples import s_pombe
    from neet.boolean.wtnetwork import WTNetwork

Weight/Threshold Networks
=========================
"""
import numpy as np
import networkx as nx
import re
from neet.statespace import StateSpace


class WTNetwork(object):
    """
    The WTNetwork class represents weight/threshold-based boolean networks. As
    such it is specified in terms of a matrix of edge weights (rows are target
    nodes) and a vector of node thresholds, and each node of the network is
    expected to be in either of two states ``0`` or ``1``.
    """

    def __init__(self, weights, thresholds=None, names=None, theta=None):
        """
        Construct a network from weights and thresholds.

        .. rubric:: Examples

        .. doctest:: wtnetwork

            >>> net = WTNetwork([[1,0],[1,1]])
            >>> net.size
            2
            >>> net.weights
            array([[1., 0.],
                   [1., 1.]])
            >>> net.thresholds
            array([0., 0.])

        .. doctest:: wtnetwork

            >>> net = WTNetwork([[1,0],[1,1]], [0.5,-0.5])
            >>> net.size
            2
            >>> net.weights
            array([[1., 0.],
                   [1., 1.]])
            >>> net.thresholds
            array([ 0.5, -0.5])

        .. doctest:: wtnetwork

            >>> net = WTNetwork(3)
            >>> net.size
            3
            >>> net.weights
            array([[0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.]])
            >>> net.thresholds
            array([0., 0., 0.])

        :param weights: the network weights, where: source/column -> target/row
        :param thresholds: the network thresholds
        :param names: the names of the network nodes (optional)
        :parma theta: the threshold function to use
        :raises ValueError: if ``weights`` is empty
        :raises ValueError: if ``weights`` is not a square matrix
        :raises ValueError: if ``thresholds`` is not a vector
        :raises ValueError: if ``weights`` and ``thresholds`` have different
                            dimensions
        :raises ValueError: if ``len(names)`` is not equal to the number of
                            nodes
        :raises TypeError: if ``threshold_func`` is not callable
        """
        if isinstance(weights, int):
            self.weights = np.zeros([weights, weights])
        else:
            self.weights = np.asarray(weights, dtype=np.float)

        shape = self.weights.shape
        if self.weights.ndim != 2:
            raise(ValueError("weights must be a matrix"))
        elif shape[0] != shape[1]:
            raise(ValueError("weights must be square"))

        if thresholds is None:
            self.thresholds = np.zeros(shape[1], dtype=np.float)
        else:
            self.thresholds = np.asarray(thresholds, dtype=np.float)

        self.__size = self.thresholds.size

        if isinstance(names, str):
            self.names = list(names)
        else:
            self.names = names

        if theta is None:
            self.theta = type(self).split_threshold
        elif callable(theta):
            self.theta = theta
        else:
            raise(TypeError("theta must be a function"))

        if self.thresholds.ndim != 1:
            raise(ValueError("thresholds must be a vector"))
        elif shape[0] != self.size:
            msg = "weights and thresholds have different dimensions"
            raise(ValueError(msg))
        elif self.size < 1:
            raise(ValueError("invalid network size"))
        elif names is not None and len(names) != self.size:
            msg = "either all or none of the nodes may have a name"
            raise(ValueError(msg))

        self.metadata = {}

    @property
    def size(self):
        """
        The number of nodes in the network.

        .. doctest:: wtnetwork

            >>> net = WTNetwork(5)
            >>> net.size
            5
            >>> WTNetwork(0)
            Traceback (most recent call last):
                ...
            ValueError: invalid network size

        :type: int
        """
        return self.__size

    def state_space(self):
        """
        Return a :class:`neet.statespace.StateSpace` object for the network.

        .. doctest:: wtnetwork

            >>> net = WTNetwork(3)
            >>> net.state_space()
            <neet.statespace.StateSpace object at 0x...>
            >>> space = net.state_space()
            >>> list(space)
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]

        :returns: the network's :class:`neet.statespace.StateSpace`
        """
        return StateSpace(self.size, base=2)

    def _unsafe_update(self, states, index=None, pin=None, values=None):
        """
        Update ``states``, in place, according to the network update rules
        without checking the validity of the arguments.

        .. rubric:: Basic Use

        .. doctest:: wtnetwork

            >>> s_pombe.size
            9
            >>> xs = [0,0,0,0,1,0,0,0,0]
            >>> s_pombe._unsafe_update(xs)
            [0, 0, 0, 0, 0, 0, 0, 0, 1]
            >>> s_pombe._unsafe_update(xs)
            [0, 1, 1, 1, 0, 0, 1, 0, 0]

        .. rubric:: Single-Node Update

        .. doctest:: wtnetwork

            >>> xs = [0,0,0,0,1,0,0,0,0]
            >>> net._unsafe_update(xs, index=-1)
            [0, 0, 0, 0, 1, 0, 0, 0, 1]
            >>> net._unsafe_update(xs, index=2)
            [0, 0, 1, 0, 1, 0, 0, 0, 1]
            >>> net._unsafe_update(xs, index=3)
            [0, 0, 1, 1, 1, 0, 0, 0, 1]


        .. rubric:: State Pinning

        .. doctest:: wtnetwork

            >>> net._unsafe_update([0,0,0,0,1,0,0,0,0], pin=[-1])
            [0, 0, 0, 0, 0, 0, 0, 0, 0]
            >>> net._unsafe_update([0,0,0,0,0,0,0,0,1], pin=[1])
            [0, 0, 1, 1, 0, 0, 1, 0, 0]
            >>> net._unsafe_update([0,0,0,0,0,0,0,0,1], pin=range(1,4))
            [0, 0, 0, 0, 0, 0, 1, 0, 0]
            >>> net._unsafe_update([0,0,0,0,0,0,0,0,1], pin=[1,2,3,-1])
            [0, 0, 0, 0, 0, 0, 1, 0, 1]

        .. rubric:: Value Fixing

        .. doctest:: wtnetwork

            >>> net.update([0,0,0,0,1,0,0,0,0], values={0:1, 2:1})
            [1, 0, 1, 0, 0, 0, 0, 0, 1]
            >>> net.update([0,0,0,0,0,0,0,0,1], values={0:1, 1:0, 2:0})
            [1, 0, 0, 1, 0, 0, 1, 0, 0]
            >>> net.update([0,0,0,0,0,0,0,0,1], values={-1:1, -2:1})
            [0, 1, 1, 1, 0, 0, 1, 1, 1]

        .. rubric:: Erroneous Usage

        .. doctest:: wtnetwork

            >>> net._unsafe_update([0,0,0])
            Traceback (most recent call last):
                ...
            ValueError: shapes (9,9) and (3,) not aligned: 9 (dim 1) != 3 (dim 0)
            >>> net._unsafe_update([0,0,0,0,2,0,0,0,0])
            [0, 0, 0, 0, 0, 0, 0, 0, 1]
            >>> net._unsafe_update([0,0,0,0,1,0,0,0,0], 9)
            Traceback (most recent call last):
                ...
            IndexError: index 9 is out of bounds for axis 0 with size 9
            >>> net._unsafe_update([0,0,0,0,0,0,0,0,1], pin=[10])
            Traceback (most recent call last):
                ...
            IndexError: index 10 is out of bounds for axis 1 with size 9

        :param states: the one-dimensional sequence of node states
        :param index: the index to update or None
        :param pin: the indices to pin (fix to their current state) or None
        :param values: a dictionary of index-value pairs to fix after update
        :returns: the updated states
        """
        pin_states = pin is not None and pin != []
        if index is None:
            if pin_states:
                pinned = np.asarray(states)[pin]
            temp = np.dot(self.weights, states) - self.thresholds
            self.theta(temp, states)
            if pin_states:
                for (j, i) in enumerate(pin):
                    states[i] = pinned[j]
        else:
            temp = np.dot(self.weights[index], states) - self.thresholds[index]
            states[index] = self.theta(temp, states[index])
        if values is not None:
            for key in values:
                states[key] = values[key]
        return states

    def update(self, states, index=None, pin=None, values=None):
        """
        Update ``states``, in place, according to the network update rules.

        .. rubric:: Basic Use

        .. doctest:: wtnetwork

            >>> s_pombe.size
            9
            >>> xs = [0,0,0,0,1,0,0,0,0]
            >>> s_pombe.update(xs)
            [0, 0, 0, 0, 0, 0, 0, 0, 1]
            >>> s_pombe.update(xs)
            [0, 1, 1, 1, 0, 0, 1, 0, 0]

        .. rubric:: Single-Node Update

        .. doctest:: wtnetwork

            >>> xs = [0,0,0,0,1,0,0,0,0]
            >>> s_pombe.update(xs, index=-1)
            [0, 0, 0, 0, 1, 0, 0, 0, 1]
            >>> s_pombe.update(xs, index=2)
            [0, 0, 1, 0, 1, 0, 0, 0, 1]
            >>> s_pombe.update(xs, index=3)
            [0, 0, 1, 1, 1, 0, 0, 0, 1]

        .. rubric:: State Pinning

        .. doctest:: wtnetwork

            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], pin=[-1])
            [0, 0, 0, 0, 0, 0, 0, 0, 0]
            >>> s_pombe.update([0,0,0,0,0,0,0,0,1], pin=[1])
            [0, 0, 1, 1, 0, 0, 1, 0, 0]
            >>> s_pombe.update([0,0,0,0,0,0,0,0,1], pin=range(1,4))
            [0, 0, 0, 0, 0, 0, 1, 0, 0]
            >>> s_pombe.update([0,0,0,0,0,0,0,0,1], pin=[1,2,3,-1])
            [0, 0, 0, 0, 0, 0, 1, 0, 1]

        .. rubric:: Value Fixing

        .. doctest:: wtnetwork

            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], values={0:1, 2:1})
            [1, 0, 1, 0, 0, 0, 0, 0, 1]
            >>> s_pombe.update([0,0,0,0,0,0,0,0,1], values={0:1, 1:0, 2:0})
            [1, 0, 0, 1, 0, 0, 1, 0, 0]
            >>> s_pombe.update([0,0,0,0,0,0,0,0,1], values={-1:1, -2:1})
            [0, 1, 1, 1, 0, 0, 1, 1, 1]

        .. rubric:: Erroneous Usage

        .. doctest:: wtnetwork

            >>> s_pombe.update([0,0,0])
            Traceback (most recent call last):
                ...
            ValueError: incorrect number of states in array
            >>> s_pombe.update([0,0,0,0,2,0,0,0,0])
            Traceback (most recent call last):
                ...
            ValueError: invalid node state in states
            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], 9)
            Traceback (most recent call last):
                ...
            IndexError: index 9 is out of bounds for axis 0 with size 9
            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], index=-1, pin=[-1])
            Traceback (most recent call last):
                ...
            ValueError: cannot provide both the index and pin arguments
            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], pin=[10])
            Traceback (most recent call last):
                ...
            IndexError: index 10 is out of bounds for axis 1 with size 9
            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], index=1, values={1:0,3:0,2:1})
            Traceback (most recent call last):
                ...
            ValueError: cannot provide both the index and values arguments
            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], pin=[1], values={1:0,3:0,2:1})
            Traceback (most recent call last):
                ...
            ValueError: cannot set a value for a pinned state
            >>> s_pombe.update([0,0,0,0,1,0,0,0,0], values={1:2})
            Traceback (most recent call last):
                ...
            ValueError: invalid state in values argument

        :param states: the one-dimensional sequence of node states
        :param index: the index to update (or None)
        :param pin: the indices to pin (or None)
        :param values: a dictionary of index-value pairs to set after update
        :returns: the updated states
        :raises ValueError: if ``states`` is not in the network's state space
        :raises ValueError: if ``index`` and ``pin`` are both provided
        :raises ValueError: if ``index`` and ``values`` are both provided
        :raises ValueError: if an element of ``pin`` is a key in ``values``
        :raises ValueError: if a value in ``values`` is not binary (0 or 1)
        """
        if states not in self.state_space():
            raise ValueError(
                "the provided state is not in the network's state space")

        if index is not None:
            if pin is not None and pin != []:
                raise ValueError(
                    "cannot provide both the index and pin arguments")
            elif values is not None and values != {}:
                raise ValueError(
                    "cannot provide both the index and values arguments")
        elif pin is not None and values is not None:
            for k in values.keys():
                if k in pin:
                    raise ValueError("cannot set a value for a pinned state")
        if values is not None:
            for val in values.values():
                if val != 0 and val != 1:
                    raise ValueError("invalid state in values argument")

        return self._unsafe_update(states, index, pin, values)

    @staticmethod
    def read(nodes_path, edges_path):
        """
        Read a network from a pair of node/edge files.

        .. doctest:: wtnetwork

            >>> nodes_path = '../neet/boolean/data/s_pombe-nodes.txt'
            >>> edges_path = '../neet/boolean/data/s_pombe-edges.txt'
            >>> net = WTNetwork.read(nodes_path, edges_path)
            >>> net.size
            9
            >>> net.names
            ['SK', 'Cdc2_Cdc13', 'Ste9', 'Rum1', 'Slp1', 'Cdc2_Cdc13_active', 'Wee1_Mik1', 'Cdc25', 'PP']

        :param nodes_path: path to the nodes file
        :type nodes_path: str
        :param edges_path: path to the edges file
        :type edges_path: str
        :returns: a :class:`WTNetwork`
        """
        comment = re.compile(r'^\s*#.*$')
        names, thresholds = [], []
        nameindices, index = dict(), 0
        with open(nodes_path, "r") as f:
            for line in f.readlines():
                if comment.match(line) is None:
                    name, threshold = line.strip().split()
                    names.append(name)
                    nameindices[name] = index
                    thresholds.append(float(threshold))
                    index += 1

        n = len(names)
        weights = np.zeros((n, n), dtype=np.float)
        with open(edges_path, "r") as f:
            for line in f.readlines():
                if comment.match(line) is None:
                    a, b, w = line.strip().split()
                    weights[nameindices[b], nameindices[a]] = float(w)

        return WTNetwork(weights, thresholds, names)

    @staticmethod
    def split_threshold(values, states):
        """
        Applies the following functional form to the arguments:

        .. math::

            \\theta_s(x,y) = \\begin{cases}
                0 & x < 0 \\\\
                y & x = 0 \\\\
                1 & x > 0.
            \\end{cases}

        If ``values`` and ``states`` are iterable, then apply the above
        function to each pair ``(x,y) in zip(values, states)`` and stores
        the result in ``states``.

        If ``values`` and ``states`` are scalar values, then simply apply
        the above threshold function to the pair ``(values, states)`` and
        return the result.

        .. rubric:: Examples

        .. doctest:: wtnetwork

            >>> ys = [0,0,0]
            >>> WTNetwork.split_threshold([1, -1, 0], ys)
            [1, 0, 0]
            >>> ys
            [1, 0, 0]
            >>> ys = [1,1,1]
            >>> WTNetwork.split_threshold([1, -1, 0], ys)
            [1, 0, 1]
            >>> ys
            [1, 0, 1]
            >>> WTNetwork.split_threshold(0,0)
            0
            >>> WTNetwork.split_threshold(0,1)
            1
            >>> WTNetwork.split_threshold(1,0)
            1
            >>> WTNetwork.split_threshold(1,1)
            1

        :param values: the threshold-shifted values of each node
        :param states: the pre-updated states of the nodes
        :returns: the updated states
        """
        if isinstance(values, list) or isinstance(values, np.ndarray):
            for i, x in enumerate(values):
                if x < 0:
                    states[i] = 0
                elif x > 0:
                    states[i] = 1
            return states
        else:
            if values < 0:
                return 0
            elif values > 0:
                return 1
            return states

    @staticmethod
    def negative_threshold(values, states):
        """
        Applies the following functional form to the arguments:

        .. math::

            \\theta_n(x) = \\begin{cases}
                0 & x \\leq 0 \\\\
                1 & x > 0.
            \\end{cases}

        If ``values`` and ``states`` are iterable, then apply the above
        function to each pair ``(x,y) in zip(values, states)`` and stores
        the result in ``states``.

        If ``values`` and ``states`` are scalar values, then simply apply
        the above threshold function to the pair ``(values, states)`` and
        return the result.

        .. rubric:: Examples

        .. doctest:: wtnetwork

            >>> ys = [0,0,0]
            >>> WTNetwork.negative_threshold([1, -1, 0], ys)
            [1, 0, 0]
            >>> ys
            [1, 0, 0]
            >>> ys = [1,1,1]
            >>> WTNetwork.negative_threshold([1, -1, 0], ys)
            [1, 0, 0]
            >>> ys
            [1, 0, 0]
            >>> WTNetwork.negative_threshold(0,0)
            0
            >>> WTNetwork.negative_threshold(0,1)
            0
            >>> WTNetwork.negative_threshold(1,0)
            1
            >>> WTNetwork.negative_threshold(1,1)
            1

        :param values: the threshold-shifted values of each node
        :param states: the pre-updated states of the nodes
        :returns: the updated states
        """
        if isinstance(values, list) or isinstance(values, np.ndarray):
            for i, x in enumerate(values):
                if x <= 0:
                    states[i] = 0
                else:
                    states[i] = 1
            return states
        else:
            if values <= 0:
                return 0
            else:
                return 1

    @staticmethod
    def positive_threshold(values, states):
        """
        Applies the following functional form to the arguments:

        .. math::

            \\theta_p(x) = \\begin{cases}
                0 & x < 0 \\\\
                1 & x \\geq 0.
            \\end{cases}

        If ``values`` and ``states`` are iterable, then apply the above
        function to each pair ``(x,y) in zip(values, states)`` and stores
        the result in ``states``.

        If ``values`` and ``states`` are scalar values, then simply apply
        the above threshold function to the pair ``(values, states)`` and
        return the result.

        .. rubric:: Examples

        .. doctest:: wtnetwork

            >>> ys = [0,0,0]
            >>> WTNetwork.positive_threshold([1, -1, 0], ys)
            [1, 0, 1]
            >>> ys
            [1, 0, 1]
            >>> ys = [1,1,1]
            >>> WTNetwork.positive_threshold([1, -1, 0], ys)
            [1, 0, 1]
            >>> ys
            [1, 0, 1]
            >>> WTNetwork.positive_threshold(0,0)
            1
            >>> WTNetwork.positive_threshold(0,1)
            1
            >>> WTNetwork.positive_threshold(1,0)
            1
            >>> WTNetwork.positive_threshold(-1,0)
            0

        :param values: the threshold-shifted values of each node
        :param states: the pre-updated states of the nodes
        :returns: the updated states
        """
        if isinstance(values, list) or isinstance(values, np.ndarray):
            for i, x in enumerate(values):
                if x < 0:
                    states[i] = 0
                else:
                    states[i] = 1
            return states
        else:
            if values < 0:
                return 0
            else:
                return 1

    def neighbors_in(self, index):
        """
        Return the set of all neighbor nodes, where edge(neighbor_node-->index)
        exists. An important consideration is that some threshold functions
        can introduce implicit dependence between nodes, e.g.
        :meth:`WTNetwork.split_threshold`.

        :param index: node index
        :returns: the set of all node indices which point toward the index node

        .. rubric:: Examples

        .. doctest:: wtnetwork

            >>> net = WTNetwork([[0,0,0],[1,0,1],[0,1,0]],
            ... theta=WTNetwork.split_threshold)
            >>> [net.neighbors_in(node) for node in range(net.size)]
            [{0}, {0, 1, 2}, {1, 2}]
            >>> net.theta = WTNetwork.negative_threshold
            >>> [net.neighbors_in(node) for node in range(net.size)]
            [set(), {0, 2}, {1}]
        """
        negative_thresh = type(self).negative_threshold
        positive_thresh = type(self).positive_threshold
        if self.theta is negative_thresh or self.theta is positive_thresh:
            return set(np.flatnonzero(self.weights[index]))
        else:
            # Assume every other theta has self loops. This will be depreciated
            # when we convert all WTNetworks to logicnetworks by default.
            return set(np.flatnonzero(self.weights[index])) | set([index])

    def neighbors_out(self, index):
        """
        Return the set of all neighbor nodes, where
        edge(index-->neighbor_node) exists.

        :param index: node index
        :returns: the set of all node indices which the index node points to

        .. rubric:: Basic Use

        .. doctest:: wtnetwork

            >>> net = WTNetwork([[0,0,0],[1,0,1],[0,1,0]],
            ... theta=WTNetwork.split_threshold)
            >>> [net.neighbors_out(node) for node in range(net.size)]
            [{0, 1}, {1, 2}, {1, 2}]
            >>> net.theta = WTNetwork.negative_threshold
            >>> [net.neighbors_out(node) for node in range(net.size)]
            [{1}, {2}, {1}]
        """
        negative_thresh = type(self).negative_threshold
        positive_thresh = type(self).positive_threshold
        if self.theta is negative_thresh or self.theta is positive_thresh:
            return set(np.flatnonzero(self.weights[:, index]))

        else:
            # Assume every other theta has self loops. This will be depreciated
            # when we convert all WTNetworks to logicnetworks by default.
            return set(np.flatnonzero(self.weights[:, index])) | set([index])

    def neighbors(self, index):
        """
        Return a set of neighbors for a specified node, or a list of sets of
        neighbors for all nodes in the network.

        :param index: node index
        :returns: a set (if index!=None) or list of sets of neighbors of a
                  node or network or nodes

        .. doctest:: wtnetwork

            >>> net = WTNetwork([[0,0,0],[1,0,1],[0,1,0]],
            ... theta=WTNetwork.split_threshold)
            >>> [net.neighbors(node) for node in range(net.size)]
            [{0, 1}, {0, 1, 2}, {1, 2}]
            >>> net.theta = WTNetwork.negative_threshold
            >>> [net.neighbors(node) for node in range(net.size)]
            [{1}, {0, 2}, {1}]
        """
        return self.neighbors_in(index) | self.neighbors_out(index)

    def to_networkx_graph(self, labels='indices'):
        """
        Return networkx graph given neet network.
        Return a ``networkx`` graph from a :class:`WTNetwork`.

        :param labels: how nodes are labeled and thus identified in networkx
                       graph (``'names'`` or ``'indices'``)
        :returns: a ``networkx.DiGraph``
        """
        if labels == 'names':
            if hasattr(self, 'names') and (self.names is not None):
                labels = self.names
            else:
                raise ValueError("network nodes do not have names")

        elif labels == 'indices':
            labels = range(self.size)

        else:
            raise ValueError("labels must be 'names' or 'indices'")

        edges = []
        for i, label in enumerate(labels):
            for j in self.neighbors_out(i):
                edges.append((labels[i], labels[j]))

        return nx.DiGraph(edges, name=self.metadata.get('name'))

    def draw(self, labels='indices', filename=None):
        """
        Output a file with a simple network drawing.

        Requires ``networkx`` and ``pygraphviz``.

        Supported image formats are determined by ``graphviz``. In particular,
        pdf support requires ``cairo`` and ``pango`` to be installed prior to
        ``graphviz`` installation.

        :param labels: how node is labeled and thus identified in networkx
                       graph ('names' or 'indices'), only used if network is
                       a :class:`neet.boolean.LogicNetwork` or
                       :class:`WTNetwork`
        :param filename: filename to write drawing to. Temporary filename will
                       be used if no filename provided.
        :returns: a ``pygraphviz`` network drawing

        """
        nx.nx_agraph.view_pygraphviz(self.to_networkx_graph(
            labels=labels), prog='circo', path=filename)



"""
.. currentmodule:: neet.boolean.randomnet

.. testsetup:: randomnet

    import numpy as np
    from neet.boolean.examples import *
    from neet.boolean.randomnet import *
    def neighbors_in(net):
        return [net.neighbors_in(node) for node in range(net.size)]
    def neighbors_out(net):
        return [net.neighbors_out(node) for node in range(net.size)]
    def in_degree(net):
        return [len(net.neighbors_in(node)) for node in range(net.size)]
    def out_degree(net):
        return [len(net.neighbors_out(node)) for node in range(net.size)]
    def mean_degree(net):
        return np.mean(out_degree(net))

Random networks
===============
"""
import random
import numpy as np
from neet.sensitivity import canalizing_nodes
from .logicnetwork import LogicNetwork


def random_logic(logic_net, p=0.5, connections='fixed-structure',
                 fix_external=False, make_irreducible=False,
                 fix_canalizing=False):
    """
    Return a ``LogicNetwork`` from an input ``LogicNetwork`` with a random
    logic table.

    ``connections`` decides how a node in the random network is connected to
    other nodes:

        ``'fixed-structure'``
            the random network has the same connections as the input network.

        ``'fixed-in-degree'``
            the number of connections to a node is the same as the input
            network, but the connections are randomly selected.

        ``'fixed-mean-degree'``
            the total number of edges is conserved, but edges are placed
            randomly between nodes.

        ``'free'``
            only the number of nodes is conserved, with the number of
            connections to a node chosen uniformly between 1 and the total
            number of nodes.

    ``p`` is the probability of a state of the connected nodes being present in
    the activation table. It is also equavolent to the probability of any node
    being activated. If ``p`` is a single number, it applies to all nodes.
    Otherwise ``p`` must be a sequence of numbers that match in size with the
    input network.

    .. doctest:: randomnet

        >>> net = random_logic(myeloid, connections='fixed-structure')
        >>> neighbors_in(net) == neighbors_in(myeloid)
        True
        >>> neighbors_out(net) == neighbors_out(myeloid)
        True

    .. doctest:: randomnet

        >>> net = random_logic(myeloid, connections='fixed-in-degree')
        >>> in_degree(net) == in_degree(myeloid)
        True
        >>> out_degree(net) == out_degree(myeloid)
        False

    .. doctest:: randomnet

        >>> net = random_logic(myeloid, connections='fixed-mean-degree')
        >>> mean_degree(net) == mean_degree(myeloid)
        True

    :param logic_net: a :class:`neet.boolean.LogicNetwork`
    :param p: probability that a state is present in the activation table
    :type p: number or sequence
    :param connections: ``'fixed-structure'``, ``'fixed-in-degree'``,
                        ``'fixed-mean-degree'``, or ``'free'``
    :type connections: str
    :returns: a random :class:`neet.boolean.LogicNetwork`
    """
    if not isinstance(logic_net, LogicNetwork):
        raise ValueError('object must be a LogicNetwork')

    if isinstance(p, (int, float)):
        ps = [p] * logic_net.size
    elif len(p) != logic_net.size:
        raise ValueError("p's length must match with network size")
    else:
        ps = p

    random_styles = {'fixed-structure': _random_logic_fixed_connections,
                     'fixed-in-degree': _random_logic_shuffled_connections,
                     'fixed-mean-degree': _random_logic_fixed_num_edges,
                     'free': _random_logic_free_connections}

    try:
        return random_styles[connections](logic_net, ps, fix_external,
                                          make_irreducible, fix_canalizing)
    except KeyError:
        msg = "connections must be 'fixed', 'fixed-in-degree'," \
              "'fixed-mean-degree', or 'free'"
        raise ValueError(msg)


def random_binary_states(k, p):
    """
    Return a set of binary states. Each state has length `k` and the number
    of states is `k * p` (or chosen to produce `k * p` on average if `n * p`
    is not an integer).
    """
    integer, decimal = divmod(2**k * p, 1)
    num_states = int(integer + np.random.choice(2, p=[1 - decimal, decimal]))
    state_idxs = np.random.choice(2 ** k, num_states, replace=False)

    return set('{0:0{1}b}'.format(idx, k) for idx in state_idxs)


def random_canalizing_binary_states(k, p):
    """
    Return a set of binary states that, when considered as a set of
    activating conditions, represents a canalizing function.

    Designed to sample each possible canalized function with equal
    probability.

    Each state has length `k` and the number of states is set in the
    same way as `random_binary_states`.
    """
    integer, decimal = divmod(2**k * p, 1)
    num_states = int(integer + np.random.choice(2, p=[1 - decimal, decimal]))

    # calculate values specifying which input is canalizing and how
    canalizing_input = np.random.choice(k)
    canalizing_value = np.random.choice(2)
    if num_states > 2**(k - 1):
        canalized_value = 1
    elif num_states < 2**(k - 1):
        canalized_value = 0
    elif num_states == 2**(k - 1):
        canalized_value = np.random.choice(2)

    fixed_states = _all_states_with_one_node_fixed(
        k, canalizing_input, canalizing_value)
    other_states = np.lib.arraysetops.setxor1d(np.arange(2**k),
                                               fixed_states,
                                               assume_unique=True)
    if canalized_value == 1:
        # include all fixed_states as activating conditions
        state_idxs = np.random.choice(other_states,
                                      num_states - len(fixed_states),
                                      replace=False)
        state_idxs = np.concatenate((state_idxs, np.array(fixed_states)))
    elif canalized_value == 0:
        # include none of fixed_states as activating conditions
        state_idxs = np.random.choice(other_states, num_states, replace=False)

    return set('{0:0{1}b}'.format(idx, k) for idx in state_idxs)


def _all_states_with_one_node_fixed(k, fixed_index, fixed_value, max_k=20):
    """
    (Should have length 2**(k-1).)
    """
    if k > max_k:
        raise Exception("k > max_k")
    # there may be a more efficient way to do this...
    return [idx for idx in range(2**k)
            if '{0:0{1}b}'.format(idx, k)[fixed_index] == str(fixed_value)]


def _external_nodes(logic_net):
    externals = set()
    for idx, row in enumerate(logic_net.table):
        if row[0] == (idx, ) and row[1] == {'1'}:
            externals.add(idx)
    return externals

# stolen from grn-survey.generate_variants


def _fake_connections(net):
    fakes = []
    for idx in range(net.size):
        for neighbor_in in net.neighbors_in(idx):
            if not net.is_dependent(idx, neighbor_in):
                fakes.append((idx, neighbor_in))
    return fakes


def _logic_table_row_is_irreducible(row, i, size):
    table = [((), set()) for j in range(size)]
    table[i] = row
    net = LogicNetwork(table)
    return len(_fake_connections(net)) == 0


def _logic_table_row_is_canalizing(row, i, size):
    table = [((), set()) for j in range(size)]
    table[i] = row
    net = LogicNetwork(table)
    return i in canalizing_nodes(net)


def _random_logic_fixed_connections(logic_net, ps, fix_external=False,
                                    make_irreducible=False,
                                    fix_canalizing=False,
                                    give_up_number=1000):
    """
    Return a `LogicNetwork` from an input `LogicNetwork` with a random logic
    table.

    Connections in the returned network are the same as those of the input.

    :param logic_net: a :class:LogicNetwork
    :param ps: probability that a state is present in the activation table
    :returns: a random :class:LogicNetwork
    """
    if not isinstance(logic_net, LogicNetwork):
        raise ValueError('object must be a LogicNetwork')

    externals = _external_nodes(logic_net)

    new_table = []
    for i, row in enumerate(logic_net.table):
        indices = row[0]
        if i in externals:
            conditions = row[1]
        else:
            if fix_canalizing:
                original_canalizing = _logic_table_row_is_canalizing(
                    row, i, logic_net.size)
            keep_trying = True
            number_tried = 0
            while keep_trying and (number_tried < give_up_number):
                if fix_canalizing and original_canalizing:
                    conditions = random_canalizing_binary_states(
                        len(indices), ps[i])
                else:
                    conditions = random_binary_states(len(indices), ps[i])

                number_tried += 1
                keep_trying = False
                if make_irreducible:
                    node_irreducible = _logic_table_row_is_irreducible(
                        (indices, conditions), i, logic_net.size)
                    keep_trying = not node_irreducible
                if (not keep_trying) and fix_canalizing:
                    node_canalizing = _logic_table_row_is_canalizing(
                        (indices, conditions), i, logic_net.size)
                    keep_trying = not (node_canalizing == original_canalizing)
            if number_tried >= give_up_number:
                msg = "No function out of " + str(give_up_number) + " tried" \
                      " satisfied constraints"
                raise Exception(msg)

        new_table.append((indices, conditions))

    return LogicNetwork(new_table, logic_net.names)


def _random_logic_shuffled_connections(logic_net, ps, fix_external=False,
                                       make_irreducible=False,
                                       fix_canalizing=False,
                                       give_up_number=1000):
    """
    Return a `LogicNetwork` from an input `LogicNetwork` with a random logic
    table.

    The number of connections to a node is the same as the input network, but
    the connections are randomly selected.

    :param logic_net: a :class:LogicNetwork
    :param p: probability that a state is present in the activation table
    :returns: a random :class:LogicNetwork
    """
    if not isinstance(logic_net, LogicNetwork):
        raise ValueError('object must be a LogicNetwork')

    externals = _external_nodes(logic_net) if fix_external else set()

    new_table = []
    for i, row in enumerate(logic_net.table):
        if i in externals:
            indices, conditions = row
        else:
            if fix_canalizing:
                original_canalizing = _logic_table_row_is_canalizing(
                    row, i, logic_net.size)
            keep_trying = True
            number_tried = 0
            while keep_trying and (number_tried < give_up_number):
                n_indices = len(row[0])
                indices = tuple(sorted(random.sample(
                    range(logic_net.size), k=n_indices)))

                if fix_canalizing and original_canalizing:
                    conditions = random_canalizing_binary_states(
                        n_indices, ps[i])
                else:
                    conditions = random_binary_states(n_indices, ps[i])

                number_tried += 1
                keep_trying = False
                if make_irreducible:
                    node_irreducible = _logic_table_row_is_irreducible(
                        (indices, conditions), i, logic_net.size)
                    keep_trying = not node_irreducible
                if (not keep_trying) and fix_canalizing:
                    node_canalizing = _logic_table_row_is_canalizing(
                        (indices, conditions), i, logic_net.size)
                    keep_trying = not (node_canalizing == original_canalizing)
            if number_tried >= give_up_number:
                msg = "No function out of " + str(give_up_number) + \
                      " tried satisfied constraints"
                raise Exception(msg)

        new_table.append((indices, conditions))

    return LogicNetwork(new_table, logic_net.names)


def _random_logic_free_connections(logic_net, ps):
    """
    Return a `LogicNetwork` from an input `LogicNetwork` with a random logic
    table.

    All possible connections within the network are considered in the random
    process.

    :param logic_net: a :class:LogicNetwork
    :param p: probability that a state is present in the activation table
    :returns: a random :class:LogicNetwork
    """
    if not isinstance(logic_net, LogicNetwork):
        raise ValueError('object must be a LogicNetwork')

    new_table = []
    for i in range(logic_net.size):
        n_indices = random.randint(1, logic_net.size)
        indices = tuple(sorted(random.sample(
            range(logic_net.size), k=n_indices)))

        conditions = random_binary_states(n_indices, ps[i])

        new_table.append((indices, conditions))

    return LogicNetwork(new_table, logic_net.names)


def _random_logic_fixed_num_edges(logic_net, ps, fix_external=False,
                                  make_irreducible=False,
                                  fix_canalizing=False,
                                  give_up_number=1000):
    """
    Returns new network that corresponds to adding a fixed number of
    edges between random nodes, with random corresponding boolean rules.
    """
    if fix_canalizing:
        raise NotImplementedError("fix_canalizing=True not yet implemented")

    num_edges = sum(len(logic_net.neighbors_in(i))
                    for i in range(logic_net.size))

    externals = _external_nodes(logic_net) if fix_external else set()

    num_edges -= len(externals)

    internals = [idx for idx in range(logic_net.size) if idx not in externals]
    num_internal_connections = np.zeros(len(internals))

    # partition edges among nodes
    keep_trying = True
    number_tried = 0
    while keep_trying and (number_tried < give_up_number):
        rng = range(len(internals) * logic_net.size)
        options = [i // logic_net.size for i in rng]
        sample = np.random.choice(options, num_edges - len(internals),
                                  replace=False)
        idxs, counts = np.unique(sample, return_counts=True)

        num_internal_connections[idxs] = counts
        num_internal_connections += 1

        number_tried += 1
        # we need to check that there is no node that will want
        # more connections than there are nodes
        if max(num_internal_connections) <= logic_net.size:
            keep_trying = False
    if number_tried >= give_up_number:
        raise Exception("No partition out of " +
                        str(give_up_number) + " tried satisfied constraints")

    new_table = [()] * logic_net.size
    for internal, num in zip(internals, num_internal_connections):
        keep_trying = True
        number_tried = 0
        while keep_trying and (number_tried < give_up_number):
            in_indices = tuple(np.random.choice(
                logic_net.size, int(num), replace=False))
            conditions = random_binary_states(len(in_indices), ps[internal])
            new_table[internal] = (in_indices, conditions)

            number_tried += 1
            if make_irreducible:
                node_irreducible = _logic_table_row_is_irreducible(
                    (in_indices, conditions), internal, logic_net.size)
                keep_trying = not node_irreducible
            else:
                keep_trying = False
        if number_tried >= give_up_number:
            raise Exception("No function out of " +
                            str(give_up_number) +
                            " tried satisfied constraints")

    for external in externals:
        new_table[external] = logic_net.table[external]

    return LogicNetwork(new_table, logic_net.names)



import unittest
from neet.interfaces import (is_network, is_fixed_sized,
                             is_boolean_network, neighbors,
                             to_networkx_graph)
from neet.statespace import StateSpace
import neet.automata as ca
import neet.boolean as bnet
from neet.boolean.examples import s_pombe


class TestCore(unittest.TestCase):
    class IsNetwork(object):
        def update(self, lattice):
            pass

        def state_space(self):
            return StateSpace(1)

        def neighbors(self):
            pass

    class FixedSizeNetwork(IsNetwork):
        def size(self):
            return 5

    class IsNotNetwork(object):
        pass

    class NotFixedSizedNetwork(IsNotNetwork):
        def size(self):
            return 5

    class BaseThreeNetwork(object):
        def update(self, lattice):
            pass

        def state_space(self):
            return StateSpace(1, base=3)

        def neighbors(self):
            pass

    class MultipleBaseNetwork(object):
        def update(self, lattice):
            pass

        def state_space(self):
            return StateSpace([1, 2, 3])

        def neighbors(self):
            pass

    def test_is_network(self):
        net = self.IsNetwork()
        self.assertTrue(is_network(net))
        self.assertTrue(is_network(type(net)))

        not_net = self.IsNotNetwork()
        self.assertFalse(is_network(not_net))
        self.assertFalse(is_network(type(not_net)))

        self.assertFalse(is_network(5))
        self.assertFalse(is_network(int))

    def test_is_fixed_sized(self):
        net = self.IsNetwork()
        self.assertFalse(is_fixed_sized(net))
        self.assertFalse(is_fixed_sized(type(net)))

        not_net = self.IsNotNetwork()
        self.assertFalse(is_fixed_sized(not_net))
        self.assertFalse(is_fixed_sized(type(not_net)))

        net = self.FixedSizeNetwork()
        self.assertTrue(is_fixed_sized(net))
        self.assertTrue(is_fixed_sized(type(net)))

        not_net = self.NotFixedSizedNetwork()
        self.assertFalse(is_fixed_sized(not_net))
        self.assertFalse(is_fixed_sized(type(not_net)))

    def test_is_boolean_network(self):
        net = self.IsNetwork()
        self.assertTrue(is_boolean_network(net))

        not_bool_net = self.BaseThreeNetwork()
        self.assertFalse(is_boolean_network(not_bool_net))

        not_bool_net = self.MultipleBaseNetwork()
        self.assertFalse(is_boolean_network(not_bool_net))

    def test_neighbors_ECA(self):
        eca = ca.ECA(30)

        with self.assertRaises(ValueError):
            neighbors(eca, 1, direction='')

        self.assertTrue(neighbors(eca, 1, size=4), set([0, 1, 2]))

        with self.assertRaises(AttributeError):
            neighbors(eca, 1)

    def test_neighbors_WTNetwork(self):
        net = bnet.WTNetwork([[1, 0], [1, 1]])

        with self.assertRaises(ValueError):
            neighbors(net, 0, direction='')

        self.assertTrue(neighbors(net, 0), [set([0])])

    def test_neighbors_LogicNetwork(self):
        net = bnet.LogicNetwork([((0,), {'0'})])

        with self.assertRaises(ValueError):
            neighbors(net, 0, direction='')

        self.assertTrue(neighbors(net, 0), [set([0])])

    def test_to_networkx_graph_LogicNetwork(self):
        net = bnet.LogicNetwork([((1, 2), {'01', '10'}),
                                 ((0, 2), ((0, 1), '10', [1, 1])),
                                 ((0, 1), {'11'})], ['A', 'B', 'C'])

        nx_net = to_networkx_graph(net, labels='names')
        self.assertEqual(set(nx_net), set(['A', 'B', 'C']))

    def test_to_networkx_graph_WTNetwork(self):

        nx_net = to_networkx_graph(s_pombe, labels='names')
        self.assertEqual(set(nx_net), set(s_pombe.names))

    def test_to_networkx_ECA_metadata(self):
        net = ca.ECA(30)
        net.boundary = (1, 0)

        with self.assertRaises(AttributeError):
            to_networkx_graph(net)

        nx_net = to_networkx_graph(net, 3)

        self.assertEqual(nx_net.graph['code'], 30)
        self.assertEqual(nx_net.graph['size'], 3)
        self.assertEqual(nx_net.graph['boundary'], (1, 0))



"""
.. currentmodule:: neet.automata

.. testsetup:: automata

    from neet.automata import *
    from neet.synchronous import transitions

Cellular Automata
=================

Cellular automata are a special type of network. They can be envisioned
at a boolean networks wherein every node has exactly 3 incoming edges. The
:mod:`neet.automata` modules provides two submodules, :mod:`neet.automata.eca`
and :mod:`neet.automata.reca`. As a convenience, the classes in each are
exposed in the :mod:`neet.automata`, so you never have to reference the
submodules unless you so choose.

API Documentation
-----------------
"""
from .eca import ECA
from .reca import RewiredECA



"""
.. currentmodule:: neet.synchronous

.. testsetup:: synchronous

    from math import e
    from neet.automata import ECA
    from neet.boolean import WTNetwork
    from neet.boolean.examples import s_pombe
    from neet.synchronous import *

Synchronous Network Analysis
============================

The :mod:`neet.synchronous` module provides a host of functions for computing
properties of the dynamics of *synchronously* updated networks. This includes

- :func:`trajectory`
- :func:`transitions`
- :func:`transition_graph`
- :func:`attractors`
- :func:`basins`
- :func:`basin_entropy`
- :func:`timeseries`

The disadvantage of these functions is that they cannot share any
information. If you wish to compute a number of them, then many values will be
repeatedly computed. For this reason, we created the :class:`Landscape` class
to simultaneously compute and several of these properties. This provides much
better performance overall.

API Documentation
-----------------
"""
import copy
import networkx as nx
import numpy as np
import pyinform as pi
from .statespace import StateSpace
from .interfaces import is_network, is_fixed_sized


def trajectory(net, state, timesteps=1, encode=False):
    """
    Generate the trajectory of length ``timesteps+1`` through the state-space,
    as determined by the network rule, beginning at ``state``.

    .. rubric:: Examples

    .. doctest:: synchronous

        >>> trajectory(ECA(30), [0, 1, 0], timesteps=3)
        [[0, 1, 0], [1, 1, 1], [0, 0, 0], [0, 0, 0]]
        >>> trajectory(ECA(30), [0, 1, 0], timesteps=3, encode=True)
        [2, 7, 0, 0]
        >>> trajectory(s_pombe, [0, 0, 0, 0, 1, 0, 0, 0, 0], timesteps=3,
        ... encode=True)
        [16, 256, 78, 128]

    :param net: the network
    :param state: the network state
    :param timesteps: the number of steps in the trajectory
    :param encode: encode the states as integers
    :returns: the trajectory as a list
    :raises TypeError: if net is not a network
    :raises ValueError: if ``timesteps < 1``
    """
    if not is_network(net):
        raise TypeError("net is not a network")
    if timesteps < 1:
        raise ValueError("number of steps must be positive, non-zero")

    traj = []
    state = copy.copy(state)
    if encode:
        if is_fixed_sized(net):
            state_space = net.state_space()
        else:
            state_space = net.state_space(len(state))

        traj.append(state_space._unsafe_encode(state))

        net.update(state)
        traj.append(state_space._unsafe_encode(state))

        for _ in range(1, timesteps):
            net._unsafe_update(state)
            traj.append(state_space._unsafe_encode(state))
    else:
        traj.append(copy.copy(state))

        net.update(state)
        traj.append(copy.copy(state))

        for _ in range(1, timesteps):
            net._unsafe_update(state)
            traj.append(copy.copy(state))
    return traj


def transitions(net, size=None, encode=False):
    """
    Generate the one-step state transitions for a network over its state space.

    .. rubric:: Examples

    .. doctest:: synchronous

        >>> transitions(ECA(30), size=3)
        [[0, 0, 0], [1, 1, 1], [1, 1, 1], [1, 0, 0], [1, 1, 1], [0, 0, 1], [0, 1, 0], [0, 0, 0]]
        >>> transitions(ECA(30), size=3, encode=True)
        [0, 7, 7, 1, 7, 4, 2, 0]
        >>> len(transitions(s_pombe, encode=True))
        512

    :param net: the network
    :param size: the size of the network (``None`` if fixed sized)
    :param encode: encode the states as integers
    :returns: the one-state transitions as an array
    :raises TypeError: if ``net`` is not a network
    :raises ValueError: if ``net`` is fixed sized and ``size`` is not ``None``
    :raises ValueError: if ``net`` is not fixed sized and ``size`` is ``None``
    """
    if not is_network(net):
        raise TypeError("net is not a network")

    if is_fixed_sized(net):
        if size is not None:
            raise ValueError("size must be None for fixed sized networks")
        state_space = net.state_space()
    else:
        if size is None:
            raise ValueError(
                "size must not be None for variable sized networks")
        state_space = net.state_space(size)

    trans = []
    for state in state_space:
        net._unsafe_update(state)
        if encode:
            trans.append(state_space._unsafe_encode(state))
        else:
            trans.append(state)

    return trans


def transition_graph(net, size=None):
    """
    Construct the state transition graph for the network.

    .. rubric:: Examples

    .. doctest:: synchronous

        >>> g = transition_graph(s_pombe)
        >>> g.number_of_nodes(), g.number_of_edges()
        (512, 512)
        >>> g = transition_graph(ECA(30), size=6)
        >>> g.number_of_nodes(), g.number_of_edges()
        (64, 64)

    :param net: the network (if already a networkx.DiGraph, does nothing and
                returns it)
    :param size: the size of the network (``None`` if fixed sized)
    :param encode: encode the states as integers
    :returns: a ``networkx.DiGraph`` of the network's transition graph
    :raises TypeError: if ``net`` is not a network
    :raises ValueError: if ``net`` is fixed sized and ``size`` is not ``None``
    :raises ValueError: if ``net`` is not fixed sized and ``size`` is ``None``
    """
    if is_network(net):
        edge_list = enumerate(transitions(net, size=size, encode=True))
        return nx.DiGraph(list(edge_list))
    elif isinstance(net, nx.DiGraph):
        if size is not None:
            raise ValueError("size must be None for transition graphs")
        return net
    else:
        raise TypeError("net must be a network or a networkx DiGraph")


def attractors(net, size=None):
    """
    Find the attractor states of a network. A generator of the attractors is
    returned with each attractor represented as a ``list`` of "encoded" states.

    .. rubric:: Examples

    .. doctest:: synchronous

        >>> attractors(s_pombe)
        [[76], [4], [8], [12], [144, 110, 384], [68], [72], [132], [136], [140], [196], [200], [204]]
        >>> attractors(ECA(30), size=5)
        [[0], [14, 25, 7, 28, 19]]

    :param net: the network or the transition graph
    :param size: the size of the network (``None`` if fixed sized)
    :returns: a list of attractor cycles
    :raises TypeError: if ``net`` is not a network or a ``networkx.DiGraph``
    :raises ValueError: if ``net`` is fixed sized and ``size`` is not ``None``
    :raises ValueError: if ``net`` is a transition graph and ``size`` is not
                        ``None``
    :raises ValueError: if ``net`` is not fixed sized and ``size`` is ``None``
    """
    if isinstance(net, nx.DiGraph):
        return list(nx.simple_cycles(net))
    elif not is_network(net):
        raise TypeError("net must be a network or a networkx DiGraph")
    elif is_fixed_sized(net) and size is not None:
        raise ValueError("fixed sized networks require size is None")
    elif not is_fixed_sized(net) and size is None:
        raise ValueError("variable sized networks require a size")
    else:
        cycles = []
        # Get the state transitions
        # (array of next state indexed by current state)
        trans = list(transitions(net, size=size, encode=True))
        # Create an array to store whether a given state has visited
        visited = np.zeros(len(trans), dtype=np.bool)
        # Create an array to store which attractor basin each state is in
        basins = np.zeros(len(trans), dtype=np.int)
        # Create a counter to keep track of how many basins have been visited
        basin_number = 1

        # Start at state 0
        initial_state = 0
        # While the initial state is a state of the system
        while initial_state < len(trans):
            # Create a stack to store the state so far visited
            state_stack = []
            # Create a array to store the states in the attractor cycle
            cycle = []
            # Create a flag to signify whether the current state is part of
            # the cycle
            in_cycle = False
            # Set the current state to the initial state
            state = initial_state
            # Store the next state and terminus variables to the next state
            terminus = next_state = trans[state]
            # Set the visited flag of the current state
            visited[state] = True
            # While the next state hasn't been visited
            while not visited[next_state]:
                # Push the current state onto the stack
                state_stack.append(state)
                # Set the current state to the next state
                state = next_state
                # Update the terminus and next_state variables
                terminus = next_state = trans[state]
                # Update the visited flag for the current state
                visited[state] = True

            # If the next state hasn't been assigned a basin yet
            if basins[next_state] == 0:
                # Set the current basin to the basin number
                basin = basin_number
                # Add the current state to the attractor cycle
                cycle.append(state)
                # We're still in the cycle until the current state is equal to
                # the terminus
                in_cycle = (terminus != state)
            else:
                # Set the current basin to the basin of next_state
                basin = basins[next_state]

            # Set the basin of the current state
            basins[state] = basin

            # While we still have states on the stack
            while len(state_stack) != 0:
                # Pop the current state off of the top of the stack
                state = state_stack.pop()
                # Set the basin of the current state
                basins[state] = basin
                # If we're still in the cycle
                if in_cycle:
                    # Add the current state to the attractor cycle
                    cycle.append(state)
                    # We're still in the cycle until the current state is
                    # equal to the terminus
                    in_cycle = (terminus != state)

            # Find the next unvisited initial state
            while initial_state < len(visited) and visited[initial_state]:
                initial_state += 1

            # Yield the cycle if we found one
            if len(cycle) != 0:
                cycles.append(cycle)
    return cycles


def basins(net, size=None):
    """
    Find the attractor basins of a network. A generator of the attractor basins
    is returned with each basin represented as a ``networkx.DiGraph`` whose
    nodes are the "encoded" network states.

    .. rubric:: Examples

    .. doctest:: synchronous

        >>> b = basins(s_pombe)
        >>> [len(basin) for basin in b]
        [378, 2, 2, 2, 104, 6, 6, 2, 2, 2, 2, 2, 2]
        >>> b = basins(ECA(30), size=5)
        >>> [len(basin) for basin in b]
        [2, 30]

    :param net: the network or landscape transition_graph
    :param size: the size of the network (``None`` if fixed sized)
    :returns: generator of basin subgraphs
    :raises TypeError: if ``net`` is not a network or a ``networkx.DiGraph``
    :raises ValueError: if ``net`` is fixed sized and ``size`` is not ``None``
    :raises ValueError: if ``net`` is a transition graph and ``size`` is not
                       ``None``
    :raises ValueError: if ``net`` is not fixed sized and ``size`` is ``None``
    """
    graph = transition_graph(net, size=size)
    return nx.weakly_connected_component_subgraphs(graph)


def basin_entropy(net, size=None, base=2):
    """
    Calculate the basin entropy [Krawitz2007]_.

    .. rubric:: Examples

    .. doctest:: synchronous

        >>> basin_entropy(s_pombe)
        1.221888833884975
        >>> basin_entropy(s_pombe, base=10)
        0.36782519036626105
        >>> basin_entropy(ECA(30), size=5)
        0.3372900666170139

    :param net: the network or landscape transition_graph
    :param size: the size of the network (``None`` if fixed sized)
    :param base: base of logarithm used to calculate entropy (2 for bits)
    :returns: value of basin entropy
    :raises TypeError: if ``net`` is not a network or a ``networkx.DiGraph``
    :raises ValueError: if ``net`` is fixed sized and ``size`` is not ``None``
    :raises ValueError: if ``net`` is a transition graph and ``size`` is not
                       ``None``
    :raises ValueError: if ``net`` is not fixed sized and ``size`` is ``None``
    """
    sizes = [len(basin) for basin in basins(net, size=size)]
    d = pi.Dist(sizes)
    return pi.shannon.entropy(d, b=base)


def timeseries(net, timesteps, size=None):
    """
    Return the timeseries for the network. The result will be a :math:`3D`
    array with shape :math:`N \\times V \\times t` where :math:`N` is the
    number of nodes in the network, :math:`V` is the volume of the state space
    (total number of network states), and :math:`t` is ``timesteps + 1``.

    .. doctest:: synchronous

        >>> net = WTNetwork([[1,-1],[1,0]])
        >>> timeseries(net, 5)
        array([[[0, 0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1, 1],
                [0, 0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1, 1]],
        <BLANKLINE>
               [[0, 0, 0, 0, 0, 0],
                [0, 1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1, 1]]])

    :param net: the network
    :param timesteps: the number of timesteps in the timeseries
    :param size: the size of the network (``None`` if fixed sized)
    :return: a numpy array
    :raises TypeError: if ``net`` is not a network
    :raises ValueError: if ``net`` is fixed sized and ``size`` is not ``None``
    :raises ValueError: if ``net`` is not fixed sized and ``size`` is ``None``
    :raises ValueError: if ``timesteps < 1``
    """
    if not is_network(net):
        raise TypeError("net must be a NEET network")
    if not is_fixed_sized(net) and size is None:
        raise ValueError("network is not fixed sized; must provide a size")
    elif is_fixed_sized(net) and size is not None:
        raise ValueError("cannot provide a size with a fixed sized network")
    if timesteps < 1:
        raise ValueError("time series must have at least one timestep")

    if size is None:
        state_space = net.state_space()
    else:
        state_space = net.state_space(size)

    shape = (state_space.ndim, state_space.volume, timesteps + 1)
    series = np.empty(shape, dtype=np.int)

    trans = list(transitions(net, size=size, encode=False))
    encoded_trans = [state_space._unsafe_encode(state) for state in trans]

    for (index, init) in enumerate(state_space):
        k = index
        series[:, index, 0] = init[:]
        for time in range(1, timesteps + 1):
            series[:, index, time] = trans[k][:]
            k = encoded_trans[k]

    return series


class Landscape(StateSpace):
    """
    The ``Landscape`` class represents the structure and topology of the
    "landscape" of state transitions. That is, it is the state space
    together with information about state transitions and the topology of
    the state transition graph.
    """

    def __init__(self, net, size=None, index=None, pin=None, values=None):
        """
        Construct the landscape for a network.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> Landscape(s_pombe)
            <neet.synchronous.Landscape object at 0x...>
            >>> Landscape(ECA(30), size=5)
            <neet.synchronous.Landscape object at 0x...>

        :param net: the network
        :param size: the size of the network (``None`` if fixed sized)
        :param index: the index to update (or None)
        :param pin: the indices to pin during update (or None)
        :param values: a dictionary of index-value pairs to set after update
        :raises TypeError: if ``net`` is not a network
        :raises ValueError: if ``net`` is fixed sized and ``size`` is not
                           ``None``
        :raises ValueError: if ``net`` is not fixed sized and ``size`` is
                           ``None``
        """

        if not is_network(net):
            raise TypeError("net is not a network")
        elif is_fixed_sized(net):
            if size is not None:
                raise ValueError("size must be None for fixed sized networks")
            state_space = net.state_space()
        else:
            if size is None:
                raise ValueError(
                    "size must not be None for variable sized networks")
            state_space = net.state_space(size)

        if state_space.is_uniform:
            super(Landscape, self).__init__(state_space.ndim, state_space.base)
        else:
            super(Landscape, self).__init__(state_space.bases)

        self.__net = net
        self.__index = index
        self.__pin = pin
        self.__values = values

        self.__expounded = False
        self.__graph = None

        self.__setup()

    @property
    def network(self):
        """
        The landscape's dynamical network

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.network
            <neet.boolean.wtnetwork.WTNetwork object at 0x...>
        """
        return self.__net

    @property
    def size(self):
        """
        The number of nodes in the landscape's dynamical network

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.size
            9
        """
        return self.ndim

    @property
    def transitions(self):
        """
        The state transitions array

        The transitions array is an array, indexed by states, whose
        values are the subsequent state of the indexing state.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.transitions
            array([  2,   2, 130, 130,   4,   0, 128, 128,   8,   0, 128, 128,  12,
                     0, 128, 128, 256, 256, 384, 384, 260, 256, 384, 384, 264, 256,
                   ...
                   208, 208, 336, 336, 464, 464, 340, 336, 464, 464, 344, 336, 464,
                   464, 348, 336, 464, 464])
        """
        return self.__transitions

    @property
    def attractors(self):
        """
        The array of attractor cycles.

        Each element of the array is an array of states in an attractor
        cycle.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.attractors
            array([array([76]), array([4]), array([8]), array([12]),
                   array([144, 110, 384]), array([68]), array([72]), array([132]),
                   array([136]), array([140]), array([196]), array([200]),
                   array([204])], dtype=object)
        """
        if not self.__expounded:
            self.__expound()
        return self.__attractors

    @property
    def basins(self):
        """
        The array of basin numbers, indexed by states.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.basins
            array([ 0,  0,  0,  0,  1,  0,  0,  0,  2,  0,  0,  0,  3,  0,  0,  0,  0,
                    0,  4,  4,  0,  0,  4,  4,  0,  0,  4,  4,  0,  0,  4,  4,  4,  4,
                    ...
                    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                    0,  0])
        """
        if not self.__expounded:
            self.__expound()
        return self.__basins

    @property
    def basin_sizes(self):
        """
        The basin sizes as an array indexed by the basin number.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.basin_sizes
            array([378,   2,   2,   2, 104,   6,   6,   2,   2,   2,   2,   2,   2])
        """
        if not self.__expounded:
            self.__expound()
        return self.__basin_sizes

    @property
    def in_degrees(self):
        """
        The in-degree of each state as an array.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.in_degrees
            array([ 6,  0,  4,  0,  2,  0,  0,  0,  2,  0,  0,  0,  2,  0,  0,  0, 12,
                    0,  4,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                    ...
                    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                    0,  0])
        """
        if not self.__expounded:
            self.__expound()
        return self.__in_degrees

    @property
    def heights(self):
        """
        The height of each state as an array.

        The *height* of a state is the number of time steps from that
        state to a state in it's attractor cycle.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.heights
            array([7, 7, 6, 6, 0, 8, 6, 6, 0, 8, 6, 6, 0, 8, 6, 6, 8, 8, 1, 1, 2, 8,
                   1, 1, 2, 8, 1, 1, 2, 8, 1, 1, 2, 2, 2, 2, 9, 9, 1, 1, 9, 9, 1, 1,
                   ...
                   3, 9, 9, 9, 3, 9, 9, 9, 3, 9, 9, 9, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                   3, 3, 3, 3, 3, 3])
        """
        if not self.__expounded:
            self.__expound()
        return self.__heights

    @property
    def attractor_lengths(self):
        """
        The length of each attractor cycle as an array, indexed by the
        attractor.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.attractor_lengths
            array([1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1])
        """
        if not self.__expounded:
            self.__expound()
        return self.__attractor_lengths

    @property
    def recurrence_times(self):
        """
        The recurrence time of each state as an array.

        The *recurrence time* is the number of time steps from that
        state until a state is repeated.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.recurrence_times
            array([7, 7, 6, 6, 0, 8, 6, 6, 0, 8, 6, 6, 0, 8, 6, 6, 8, 8, 3, 3, 2, 8,
                   3, 3, 2, 8, 3, 3, 2, 8, 3, 3, 4, 4, 4, 4, 9, 9, 3, 3, 9, 9, 3, 3,
                   ...
                   3, 9, 9, 9, 3, 9, 9, 9, 3, 9, 9, 9, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                   3, 3, 3, 3, 3, 3])
        """
        if not self.__expounded:
            self.__expound()
        return self.__recurrence_times

    @property
    def graph(self):
        """
        The state transitions graph of the landscape as a
        ``networkx.Digraph``.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.graph
            <networkx.classes.digraph.DiGraph object at 0x106504810>
        """
        if self.__graph is None:
            self.__graph = nx.DiGraph(list(enumerate(self.__transitions)))
        return self.__graph

    def __setup(self):
        """
        This function performs all of the initilization-time setup of
        the ``Landscape`` object. At present this is limited to
        computing the state transitions array, but subsequent versions
        may expand the work that ``__setup`` does.
        """
        update = self.__net._unsafe_update
        encode = self._unsafe_encode

        transitions = np.empty(self.volume, dtype=np.int)
        for i, state in enumerate(self):
            transitions[i] = encode(update(state,
                                           index=self.__index,
                                           pin=self.__pin,
                                           values=self.__values))

        self.__transitions = transitions

    def __expound(self):
        """
        This function performs the bulk of the calculations that the
        ``Landscape`` is concerned with. Most of the properties in this
        class are computed by this function whenever *any one* of them
        is requested and the results are cached. The advantage of this
        is that it saves computation time; why traverse the state space
        for every property call when you can do it all at once. The
        downside is that the cached results may use a good bit more
        memory. This is a trade-off that we are willing to make for now.

        The properties that are computed by this function include:
            - :method:`attractors`
            - :method:`basins`
            - :method:`basin_sizes`
            - :method:`in_degrees`
            - :method:`heights`
            - :method:`attractor_lengths`
            - :method:`recurrence_times`
        """
        # Get the state transitions
        trans = self.__transitions
        # Create an array to store whether a given state has visited
        visited = np.zeros(self.volume, dtype=np.bool)
        # Create an array to store which attractor basin each state is in
        basins = np.full(self.volume, -1, dtype=np.int)
        # Create an array to store the in-degree of each state
        in_degrees = np.zeros(self.volume, dtype=np.int)
        # Create an array to store the height of each state
        heights = np.zeros(self.volume, dtype=np.int)
        # Create an array to store the recurrence time of each state
        recurrence_times = np.zeros(self.volume, dtype=np.int)
        # Create a counter to keep track of how many basins have been visited
        basin_number = 0
        # Create a list of basin sizes
        basin_sizes = []
        # Create a list of attractor cycles
        attractors = []
        # Create a list of attractor lengths
        attractor_lengths = []

        # Start at state 0
        initial_state = 0
        # While the initial state is a state of the system
        while initial_state < len(trans):
            # Create a stack to store the state so far visited
            state_stack = []
            # Create a array to store the states in the attractor cycle
            cycle = []
            # Create a flag to signify whether the current state is part of
            # the cycle
            in_cycle = False
            # Set the current state to the initial state
            state = initial_state
            # Store the next state and terminus variables to the next state
            terminus = next_state = trans[state]
            # Set the visited flag of the current state
            visited[state] = True
            # Increment in-degree
            in_degrees[next_state] += 1
            # While the next state hasn't been visited
            while not visited[next_state]:
                # Push the current state onto the stack
                state_stack.append(state)
                # Set the current state to the next state
                state = next_state
                # Update the terminus and next_state variables
                terminus = next_state = trans[state]
                # Update the visited flag for the current state
                visited[state] = True
                # Increment in-degree
                in_degrees[next_state] += 1

            # If the next state hasn't been assigned a basin yet
            if basins[next_state] == -1:
                # Set the current basin to the basin number
                basin = basin_number
                # Increment the basin number
                basin_number += 1
                # Add a new basin size
                basin_sizes.append(0)
                # Add a new attractor length
                attractor_lengths.append(1)
                # Add the current state to the attractor cycle
                cycle.append(state)
                # Set the current state's recurrence time
                recurrence_times[state] = 0
                # We're still in the cycle until the current state is equal to
                # the terminus
                in_cycle = (terminus != state)
            else:
                # Set the current basin to the basin of next_state
                basin = basins[next_state]
                # Set the state's height to one greater than the next state's
                heights[state] = heights[next_state] + 1
                # Set the state's recurrence time to one greater than the next
                # state's
                recurrence_times[state] = recurrence_times[next_state] + 1

            # Set the basin of the current state
            basins[state] = basin
            # Increment the basin size
            basin_sizes[basin] += 1

            # While we still have states on the stack
            while len(state_stack) != 0:
                # Save the current state as the next state
                next_state = state
                # Pop the current state off of the top of the stack
                state = state_stack.pop()
                # Set the basin of the current state
                basins[state] = basin
                # Increment the basin_size
                basin_sizes[basin] += 1
                # If we're still in the cycle
                if in_cycle:
                    # Add the current state to the attractor cycle
                    cycle.append(state)
                    # Increment the current attractor length
                    attractor_lengths[basin] += 1
                    # We're still in the cycle until the current state is
                    # equal to the terminus
                    in_cycle = (terminus != state)
                    # Set the cycle state's recurrence times
                    if not in_cycle:
                        for cycle_state in cycle:
                            rec_time = attractor_lengths[basin] - 1
                            recurrence_times[cycle_state] = rec_time
                else:
                    # Set the state's height to one create than the next
                    # state's
                    heights[state] = heights[next_state] + 1
                    # Set the state's recurrence time to one greater than the
                    # next state's
                    recurrence_times[state] = recurrence_times[next_state] + 1

            # Find the next unvisited initial state
            while initial_state < len(visited) and visited[initial_state]:
                initial_state += 1

            # If the cycle isn't empty, append it to the attractors list
            if len(cycle) != 0:
                attractors.append(np.asarray(cycle, dtype=np.int))

        self.__basins = basins
        self.__basin_sizes = np.asarray(basin_sizes)
        self.__attractors = np.asarray(attractors)
        self.__attractor_lengths = np.asarray(attractor_lengths)
        self.__in_degrees = in_degrees
        self.__heights = heights
        self.__recurrence_times = np.asarray(recurrence_times)
        self.__expounded = True

    def trajectory(self, init, timesteps=None, encode=None):
        """
        Compute the trajectory of a state.

        This method computes a trajectory from ``init`` to the last
        before the trajectory begins to repeat. If ``timesteps`` is
        provided, then the trajectory will have a length of ``timesteps
        + 1`` regardless of repeated states. The ``encode`` argument
        forces the states in the trajectory to be either encoded or not.
        When ``encode is None``, whether or not the states of the
        trajectory are encoded is determined by whether or not the
        initial state (``init``) is provided in encoded form.

        Note that when ``timesteps is None``, the length of the
        resulting trajectory should be one greater than the recurrence
        time of the state.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.trajectory([1,0,0,1,0,1,1,0,1])
            [[1, 0, 0, 1, 0, 1, 1, 0, 1], ... [0, 0, 1, 1, 0, 0, 1, 0, 0]]

            >>> landscape.trajectory([1,0,0,1,0,1,1,0,1], encode=True)
            [361, 80, 320, 78, 128, 162, 178, 400, 332, 76]

            >>> landscape.trajectory(361)
            [361, 80, 320, 78, 128, 162, 178, 400, 332, 76]

            >>> landscape.trajectory(361, encode=False)
            [[1, 0, 0, 1, 0, 1, 1, 0, 1], ... [0, 0, 1, 1, 0, 0, 1, 0, 0]]

            >>> landscape.trajectory(361, timesteps=5)
            [361, 80, 320, 78, 128, 162]

            >>> landscape.trajectory(361, timesteps=10)
            [361, 80, 320, 78, 128, 162, 178, 400, 332, 76, 76]

        :param init: the initial state
        :type init: ``int`` or an iterable
        :param timesteps: the number of time steps to include in the trajectory
        :type timesteps: ``int`` or ``None``
        :param encode: whether to encode the states in the trajectory
        :type encode: ``bool`` or ``None``
        :return: a list whose elements are subsequent states of the trajectory

        :raises ValueError: if ``init`` an empty array
        :raises ValueError: if ``timesteps`` is less than :math:`1`
        """
        decoded = isinstance(init, list) or isinstance(init, np.ndarray)

        if decoded:
            if init == []:
                raise ValueError("initial state cannot be empty")
            elif encode is None:
                encode = False
            init = self.encode(init)
        elif encode is None:
            encode = True

        trans = self.__transitions
        if timesteps is not None:
            if timesteps < 1:
                raise ValueError("number of steps must be positive, non-zero")

            path = [init] * (timesteps + 1)
            for i in range(1, len(path)):
                path[i] = trans[path[i - 1]]
        else:
            path = [init]
            state = trans[init]
            while state not in path:
                path.append(state)
                state = trans[state]

        if not encode:
            decode = self.decode
            path = [decode(state) for state in path]

        return path

    def timeseries(self, timesteps):
        """
        Compute the full time series of the landscape.

        This method computes a 3-dimensional array elements are the
        states of each node in the network. The dimensions of the array
        are indexed by, in order, the node, the initial state and the
        time step.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.timeseries(5)
            array([[[0, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    ...,
                    [1, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0, 0]],
            <BLANKLINE>
                   [[0, 1, 1, 1, 1, 0],
                    [0, 1, 1, 1, 1, 0],
                    [1, 1, 1, 1, 0, 0],
                    ...,
                    [0, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0, 0]],
            <BLANKLINE>
                   ...
            <BLANKLINE>
                   [[0, 0, 1, 1, 1, 1],
                    [0, 0, 1, 1, 1, 1],
                    [0, 1, 1, 1, 1, 0],
                    ...,
                    [1, 0, 0, 0, 0, 0],
                    [1, 1, 0, 0, 0, 0],
                    [1, 1, 0, 0, 0, 0]],
            <BLANKLINE>
                   [[0, 0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 1, 1],
                    ...,
                    [1, 1, 1, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0]]])

        :param timesteps: the number of timesteps to evolve the system
        :type timesteps: ``int``
        :return: a 3-D array of node states

        :raises ValueError: if ``timesteps`` is less than :math:`1`
        """
        if timesteps < 1:
            raise ValueError("number of steps must be positive, non-zero")

        trans = self.__transitions
        decode = self.decode
        decoded_trans = [decode(state) for state in trans]

        shape = (self.ndim, self.volume, timesteps + 1)
        series = np.empty(shape, dtype=np.int)

        for index, init in enumerate(self):
            k = index
            series[:, index, 0] = init[:]
            for time in range(1, timesteps + 1):
                series[:, index, time] = decoded_trans[k][:]
                k = trans[k]

        return series

    def basin_entropy(self, base=2.0):
        """
        Compute the basin entropy of the landscape [Krawitz2007]_.

        This method computes the Shannon entropy of the distribution of
        basin sizes. The base of the logarithm is chosen to be the
        number of basins so that the result is :math:`0 \leq h \leq 1`.
        If there is fewer than :math:`2` basins, then the base is taken
        to be :math:`2` so that the result is never `NaN`. The base can
        be forcibly overridden with the ``base`` keyword argument.

        .. rubric:: Examples

        .. doctest:: synchronous

            >>> landscape = Landscape(s_pombe)
            >>> landscape.basin_entropy()
            1.221888833884975
            >>> landscape.basin_entropy(base=2)
            1.221888833884975
            >>> landscape.basin_entropy(base=10)
            0.36782519036626105
            >>> landscape.basin_entropy(base=e)
            0.8469488001650497

        :param base: the base of the logarithm
        :type base: a number or ``None``
        :return: the basin entropy of the landscape of type ``float``
        """
        if not self.__expounded:
            self.__expound()
        dist = pi.Dist(self.__basin_sizes)
        return pi.shannon.entropy(dist, b=base)



"""
The :mod:`neet.python3` provides type aliases which make it easier to write
code which works under both Python 2 and Python 3.
"""
import sys

if sys.version_info > (3,):
    long = int
    unicode = str
else:
    long = long
    unicode = unicode



import numpy as np
import unittest
from neet.automata import ECA
from neet.boolean.examples import s_pombe
from neet.information import (Architecture, active_information,
                              entropy_rate, transfer_entropy,
                              mutual_information)


class TestInformation(unittest.TestCase):
    """
    Test the information analysis module
    """

    def test_canary(self):
        """
        A canary test to ensure the test suite is working
        """
        self.assertEqual(3, 1+2)

    def test_active_info_not_network(self):
        """
        Raise a ``TypeError`` if the provided network is not actually a network
        """
        with self.assertRaises(TypeError):
            active_information(5, k=3, timesteps=10, local=False)
        with self.assertRaises(TypeError):
            active_information(5, k=3, timesteps=10, local=True)

    def test_active_info_not_fixed_size(self):
        """
        Raise a ``ValueError`` if the provided network is not fixed sized, and
        the ``size`` argument is ``None``
        """
        with self.assertRaises(ValueError):
            active_information(ECA(30), k=3, timesteps=10, local=False)
        active_information(ECA(30), k=3, timesteps=10, size=5, local=False)

    def test_active_info_s_pombe(self):
        """
        ``active_information`` computes the correct values for ``s_pombe``
        """
        known_ai = [0.0, 0.408344, 0.629567, 0.629567, 0.379157, 0.400462,
                    0.670196, 0.670196, 0.391891]
        computed_ai = active_information(s_pombe, k=5, timesteps=20)
        self.assertEqual(9, len(computed_ai))
        for got, expected in zip(computed_ai, known_ai):
            self.assertAlmostEqual(expected, got, places=6)

    def test_local_active_info_s_pombe(self):
        """
        local ``active_information`` averages to the correct values for
        ``s_pombe``
        """
        known_ai = [0.0, 0.408344, 0.629567, 0.629567, 0.379157, 0.400462,
                    0.670196, 0.670196, 0.391891]
        computed_ai = active_information(
            s_pombe, k=5, timesteps=20, local=True)
        self.assertEqual((9, 512, 16), computed_ai.shape)
        for got, expected in zip(computed_ai, known_ai):
            self.assertAlmostEqual(expected, np.mean(got), places=6)

    def test_entropy_rate_not_network(self):
        """
        Raise a ``TypeError`` if the provided network is not actually a network
        """
        with self.assertRaises(TypeError):
            entropy_rate(5, k=3, timesteps=10, local=False)
        with self.assertRaises(TypeError):
            entropy_rate(5, k=3, timesteps=10, local=True)

    def test_entropy_rate_not_fixed_size(self):
        """
        Raise a ``ValueError`` if the provided network is not fixed sized, and
        the ``size`` argument is ``None``
        """
        with self.assertRaises(ValueError):
            entropy_rate(ECA(30), k=3, timesteps=10, local=False)
        entropy_rate(ECA(30), k=3, timesteps=10, size=5, local=False)

    def test_entropy_rate_s_pombe(self):
        """
        ``entropy_rate`` computes the correct values for ``s_pombe``
        """
        known_er = [0.0, 0.016912, 0.072803, 0.072803, 0.058420, 0.024794,
                    0.032173, 0.032173, 0.089669]
        computed_er = entropy_rate(s_pombe, k=5, timesteps=20)
        self.assertEqual(9, len(computed_er))
        for got, expected in zip(computed_er, known_er):
            self.assertAlmostEqual(expected, got, places=6)

    def test_local_entropy_rate_s_pombe(self):
        """
        local ``entropy_rate`` averages to the correct values for ``s_pombe``
        """
        known_er = [0.0, 0.016912, 0.072803, 0.072803, 0.058420, 0.024794,
                    0.032173, 0.032173, 0.089669]
        computed_er = entropy_rate(s_pombe, k=5, timesteps=20, local=True)
        self.assertEqual((9, 512, 16), computed_er.shape)
        for got, expected in zip(computed_er, known_er):
            self.assertAlmostEqual(expected, np.mean(got), places=6)

    def test_transfer_entropy_not_network(self):
        """
        Raise a ``TypeError`` if the provided network is not actually a network
        """
        with self.assertRaises(TypeError):
            transfer_entropy(5, k=3, timesteps=10, local=False)
        with self.assertRaises(TypeError):
            transfer_entropy(5, k=3, timesteps=10, local=True)

    def test_transfer_entropy_not_fixed_size(self):
        """
        Raise a ``ValueError`` if the provided network is not fixed sized, and
        the ``size`` argument is ``None``
        """
        with self.assertRaises(ValueError):
            transfer_entropy(ECA(30), k=3, timesteps=10, local=False)
        transfer_entropy(ECA(30), k=3, timesteps=10, size=5, local=False)

    def test_transfer_entropy_s_pombe(self):
        """
        ``transfer_entropy`` computes the correct values for ``s_pombe``
        """
        known_te = np.asarray(
            [[0., 0., 0., 0., 0., 0., 0., 0., 0.],
             [0., 0., 0., 0., 0.016912, 0., 0., 0., 0.],
             [0., 0.051370, 0., 0.012225, 0.019947,
                 0.051370, 0.006039, 0.006039, 0.072803],
             [0., 0.051370, 0.012225, 0., 0.019947,
                 0.051370, 0.006039, 0.006039, 0.072803],
             [0., 0.058420, 0.047602, 0.047602, 0.,
                 0.058420, 0.047602, 0.047602, 0.],
             [0., 0., 0.024794, 0.024794, 0., 0., 0.024794, 0.024794, 0.],
             [0., 0.016690, 0.004526, 0.004526, 0.011916,
                 0.016690, 0., 0.002983, 0.032173],
             [0., 0.016690, 0.004526, 0.004526, 0.011916,
                 0.016690, 0.002983, 0., 0.032173],
             [0., 0.060304, 0.048289, 0.048289, 0.089669, 0.060304,
                 0.048927, 0.048927, 0.]])
        computed_te = transfer_entropy(s_pombe, k=5, timesteps=20)
        self.assertEqual(known_te.shape, computed_te.shape)
        for got, expected in zip(computed_te.flatten(), known_te.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

    def test_local_transfer_entropy_s_pombe(self):
        """
        local ``transfer_entropy`` averages to the correct values for
        ``s_pombe``
        """
        known_te = np.asarray(
            [[0., 0., 0., 0., 0., 0., 0., 0., 0.],
             [0., 0., 0., 0., 0.016912, 0., 0., 0., 0.],
             [0., 0.051370, 0., 0.012225, 0.019947,
                 0.051370, 0.006039, 0.006039, 0.072803],
             [0., 0.051370, 0.012225, 0., 0.019947,
                 0.051370, 0.006039, 0.006039, 0.072803],
             [0., 0.058420, 0.047602, 0.047602, 0.,
                 0.058420, 0.047602, 0.047602, 0.],
             [0., 0., 0.024794, 0.024794, 0., 0., 0.024794, 0.024794, 0.],
             [0., 0.016690, 0.004526, 0.004526, 0.011916,
                 0.016690, 0., 0.002983, 0.032173],
             [0., 0.016690, 0.004526, 0.004526, 0.011916,
                 0.016690, 0.002983, 0., 0.032173],
             [0., 0.060304, 0.048289, 0.048289, 0.089669, 0.060304,
                 0.048927, 0.048927, 0.]])
        computed_te = transfer_entropy(s_pombe, k=5, timesteps=20, local=True)
        self.assertEqual((9, 9, 512, 16), computed_te.shape)
        for i in range(9):
            for j in range(9):
                self.assertAlmostEqual(
                    known_te[i, j], np.mean(computed_te[i, j]), places=6)

    def test_mutual_information_not_network(self):
        """
        Raise a ``TypeError`` if the provided network is not actually a network
        """
        with self.assertRaises(TypeError):
            mutual_information(5, timesteps=10, local=False)
        with self.assertRaises(TypeError):
            mutual_information(5, timesteps=10, local=True)

    def test_mutual_information_not_fixed_size(self):
        """
        Raise a ``ValueError`` if the provided network is not fixed sized, and
        the ``size`` argument is ``None``
        """
        with self.assertRaises(ValueError):
            mutual_information(ECA(30), timesteps=10, local=False)
        mutual_information(ECA(30), timesteps=10, size=5, local=False)

    def test_mutual_information_s_pombe(self):
        """
        ``mutual_information`` computes the correct values for ``s_pombe``
        """
        known_mi = np.asarray(
            [[0.162326, 0.013747, 0.004285, 0.004285, 0.013409, 0.015862,
              0.005170, 0.005170, 0.011028],
             [0.013747, 0.566610, 0.007457, 0.007457, 0.006391, 0.327908,
              0.006761, 0.006761, 0.004683],
             [0.004285, 0.007457, 0.838373, 0.475582, 0.211577, 0.004329,
              0.459025, 0.459025, 0.127557],
             [0.004285, 0.007457, 0.475582, 0.838373, 0.211577, 0.004329,
              0.459025, 0.459025, 0.127557],
             [0.013409, 0.006391, 0.211577, 0.211577, 0.574591, 0.007031,
              0.175608, 0.175608, 0.012334],
             [0.015862, 0.327908, 0.004329, 0.004329, 0.007031, 0.519051,
              0.006211, 0.006211, 0.002607],
             [0.005170, 0.006761, 0.459025, 0.459025, 0.175608, 0.006211,
              0.808317, 0.493495, 0.103905],
             [0.005170, 0.006761, 0.459025, 0.459025, 0.175608, 0.006211,
              0.493495, 0.808317, 0.103905],
             [0.011028, 0.004683, 0.127557, 0.127557, 0.012334, 0.002607,
              0.103905, 0.103905, 0.634238]])
        computed_mi = mutual_information(s_pombe, timesteps=20)
        self.assertEqual(known_mi.shape, computed_mi.shape)
        for got, expected in zip(computed_mi.flatten(), known_mi.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

    def test_local_mutual_information_s_pombe(self):
        """
        local ``mutual_information`` averages to the correct values for
        ``s_pombe``
        """
        known_mi = np.asarray(
            [[0.162326, 0.013747, 0.004285, 0.004285, 0.013409, 0.015862,
              0.005170, 0.005170, 0.011028],
             [0.013747, 0.566610, 0.007457, 0.007457, 0.006391, 0.327908,
              0.006761, 0.006761, 0.004683],
             [0.004285, 0.007457, 0.838373, 0.475582, 0.211577, 0.004329,
              0.459025, 0.459025, 0.127557],
             [0.004285, 0.007457, 0.475582, 0.838373, 0.211577, 0.004329,
              0.459025, 0.459025, 0.127557],
             [0.013409, 0.006391, 0.211577, 0.211577, 0.574591, 0.007031,
              0.175608, 0.175608, 0.012334],
             [0.015862, 0.327908, 0.004329, 0.004329, 0.007031, 0.519051,
              0.006211, 0.006211, 0.002607],
             [0.005170, 0.006761, 0.459025, 0.459025, 0.175608, 0.006211,
              0.808317, 0.493495, 0.103905],
             [0.005170, 0.006761, 0.459025, 0.459025, 0.175608, 0.006211,
              0.493495, 0.808317, 0.103905],
             [0.011028, 0.004683, 0.127557, 0.127557, 0.012334, 0.002607,
              0.103905, 0.103905, 0.634238]])
        computed_mi = mutual_information(s_pombe, timesteps=20, local=True)
        self.assertEqual((9, 9, 512, 21), computed_mi.shape)
        for i in range(9):
            for j in range(9):
                self.assertAlmostEqual(
                    known_mi[i, j], np.mean(computed_mi[i, j]), places=6)

    def test_architecture_ai(self):
        """
        The architecture correctly computes the active information
        """
        k, timesteps = 5, 20
        arch = Architecture(s_pombe, k=k, timesteps=timesteps)

        expected_ai = active_information(s_pombe, k=k, timesteps=timesteps)
        got_ai = arch.active_information()
        self.assertEqual(got_ai.shape, expected_ai.shape)
        for got, expected in zip(got_ai, expected_ai):
            self.assertAlmostEqual(expected, got, places=6)

        expected_ai = active_information(
            s_pombe, k=k, timesteps=timesteps, local=True)
        got_ai = arch.active_information(local=True)
        self.assertEqual(got_ai.shape, expected_ai.shape)
        for got, expected in zip(got_ai.flatten(), expected_ai.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

    def test_architecture_er(self):
        """
        The architecture correctly computes the entropy rate
        """
        k, timesteps = 5, 20
        arch = Architecture(s_pombe, k=k, timesteps=timesteps)

        expected_er = entropy_rate(s_pombe, k=k, timesteps=timesteps)
        got_er = arch.entropy_rate()
        self.assertEqual(got_er.shape, expected_er.shape)
        for got, expected in zip(got_er, expected_er):
            self.assertAlmostEqual(expected, got, places=6)

        expected_er = entropy_rate(
            s_pombe, k=k, timesteps=timesteps, local=True)
        got_er = arch.entropy_rate(local=True)
        self.assertEqual(got_er.shape, expected_er.shape)
        for got, expected in zip(got_er.flatten(), expected_er.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

    def test_architecture_te(self):
        """
        The architecture correctly computes the transfer entropy
        """
        k, timesteps = 5, 20
        arch = Architecture(s_pombe, k=k, timesteps=timesteps)

        expected_te = transfer_entropy(s_pombe, k=k, timesteps=timesteps)
        got_te = arch.transfer_entropy()
        self.assertEqual(got_te.shape, expected_te.shape)
        for got, expected in zip(got_te.flatten(), expected_te.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

        expected_te = transfer_entropy(
            s_pombe, k=k, timesteps=timesteps, local=True)
        got_te = arch.transfer_entropy(local=True)
        self.assertEqual(got_te.shape, expected_te.shape)
        for got, expected in zip(got_te.flatten(), expected_te.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

    def test_architecture_mi(self):
        """
        The architecture correctly computes the mutual information
        """
        k, timesteps = 5, 20
        arch = Architecture(s_pombe, k=k, timesteps=timesteps)

        expected_mi = mutual_information(s_pombe, timesteps=timesteps)
        got_mi = arch.mutual_information()
        self.assertEqual(got_mi.shape, expected_mi.shape)
        for got, expected in zip(got_mi.flatten(), expected_mi.flatten()):
            self.assertAlmostEqual(expected, got, places=6)

        expected_mi = mutual_information(
            s_pombe, timesteps=timesteps, local=True)
        got_mi = arch.mutual_information(local=True)
        self.assertEqual(got_mi.shape, expected_mi.shape)
        for got, expected in zip(got_mi.flatten(), expected_mi.flatten()):
            self.assertAlmostEqual(expected, got, places=6)



"""
.. currentmodule:: neet.sensitivity

.. testsetup:: sensitivity

    from neet.boolean.examples import c_elegans, s_pombe
    from neet.sensitivity import *

Sensitivity
===========

The :mod:`neet.sensitivity` module provides a collection of functions for
computing measures of sensitivity of networks, i.e. the degree to which
perturbations of the network state propogate and spread. This module also
provides a collection of functions for identifying "canalizing edges": edges
for which a state of the source node uniquely determines the state of the
target regardless of other sources.

API Documentation
-----------------
"""
from .interfaces import is_boolean_network
from .synchronous import transitions

import copy
import numpy as np
import numpy.linalg as linalg


def sensitivity(net, state, transitions=None):
    """
    Calculate Boolean network sensitivity, as defined in [Shmulevich2004]_

    The sensitivity of a Boolean function :math:`f` on state vector :math:`x`
    is the number of Hamming neighbors of :math:`x` on which the function
    value is different than on :math:`x`.

    This calculates the average sensitivity over all :math:`N` boolean
    functions, where :math:`N` is the size of net.

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> sensitivity(s_pombe, [0, 0, 0, 0, 0, 1, 1, 0, 0])
        1.0
        >>> sensitivity(s_pombe, [0, 1, 1, 0, 1, 0, 0, 1, 0])
        0.4444444444444444
        >>> sensitivity(c_elegans, [0, 0, 0, 0, 0, 0, 0, 0])
        1.75
        >>> sensitivity(c_elegans, [1, 1, 1, 1, 1, 1, 1, 1])
        1.25

    :param net: :mod:`neet` boolean network
    :param state: a single network state, represented as a list of node states
    :param transitions: a list of precomputed state transitions (*optional*)
    :type transitions: list or None
    """
    if not is_boolean_network(net):
        raise(TypeError("net must be a boolean network"))

    # list Hamming neighbors
    neighbors = _hamming_neighbors(state)

    nextState = net.update(state)

    # count sum of differences found in neighbors of the original
    s = 0.
    for neighbor in neighbors:
        if transitions is not None:
            newState = transitions[_fast_encode(neighbor)]
        else:
            newState = net._unsafe_update(neighbor)
        s += _boolean_distance(newState, nextState)

    return s / net.size


def difference_matrix(net, state, transitions=None):
    """
    Returns matrix answering the question: Starting at the given state, does
    flipping the state of node ``j`` change the state of node ``i``?

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> difference_matrix(s_pombe, [0, 0, 0, 0, 0, 0, 0, 0, 0])
        array([[0., 0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 1., 1., 1., 0., 0., 0., 0.],
               [0., 0., 1., 0., 0., 0., 0., 0., 1.],
               [0., 0., 0., 1., 0., 0., 0., 0., 1.],
               [0., 0., 0., 0., 0., 1., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 1., 0.],
               [0., 0., 0., 0., 0., 0., 1., 0., 1.],
               [0., 1., 0., 0., 0., 0., 0., 1., 0.],
               [0., 0., 0., 0., 1., 0., 0., 0., 0.]])
        >>> difference_matrix(c_elegans, [0, 0, 0, 0, 0, 0, 0, 0])
        array([[1., 0., 0., 0., 0., 0., 0., 1.],
               [0., 0., 1., 1., 0., 0., 0., 0.],
               [0., 0., 1., 0., 1., 0., 0., 0.],
               [0., 0., 1., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 1., 1., 0., 1.],
               [0., 0., 0., 0., 0., 1., 1., 0.],
               [1., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 1.]])

    :param net: a :mod:`neet` boolean network
    :param state: the starting state
    :param transitions: a precomputed list of state transitions (*optional*)
    :type transitions: list or None
    """
    # set up empty matrix
    N = len(state)
    Q = np.empty((N, N))

    # list Hamming neighbors (in order!)
    neighbors = _hamming_neighbors(state)

    nextState = net.update(state)

    # count differences found in neighbors of the original
    for j, neighbor in enumerate(neighbors):
        if transitions is not None:
            newState = transitions[_fast_encode(neighbor)]
        else:
            newState = net._unsafe_update(neighbor)
        Q[:, j] = [(nextState[i] + newState[i]) % 2 for i in range(N)]

    return Q


def _states_limited(nodes, state):
    """
    All possible states that vary only nodes with given indices.
    """
    if len(nodes) == 0:
        return [state]
    for i in nodes:
        stateFlipped = copy.copy(state)
        stateFlipped[nodes[0]] = (stateFlipped[nodes[0]] + 1) % 2

        left_rec = _states_limited(nodes[1:], state)
        right_rec = _states_limited(nodes[1:], stateFlipped)
        return left_rec + right_rec


def average_difference_matrix(net, states=None, weights=None, calc_trans=True):
    """
    Averaged over states, what is the probability
    that node i's state is changed by a single bit flip of node j?

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> average_difference_matrix(s_pombe)
        array([[0.    , 0.    , 0.    , 0.    , 0.    , 0.    , 0.    , 0.    ,
                0.    ],
               [0.    , 0.    , 0.25  , 0.25  , 0.25  , 0.    , 0.    , 0.    ,
                0.    ],
               [0.25  , 0.25  , 0.25  , 0.    , 0.    , 0.25  , 0.    , 0.    ,
                0.25  ],
               [0.25  , 0.25  , 0.    , 0.25  , 0.    , 0.25  , 0.    , 0.    ,
                0.25  ],
               [0.    , 0.    , 0.    , 0.    , 0.    , 1.    , 0.    , 0.    ,
                0.    ],
               [0.    , 0.    , 0.0625, 0.0625, 0.0625, 0.    , 0.0625, 0.0625,
                0.    ],
               [0.    , 0.5   , 0.    , 0.    , 0.    , 0.    , 0.5   , 0.    ,
                0.5   ],
               [0.    , 0.5   , 0.    , 0.    , 0.    , 0.    , 0.    , 0.5   ,
                0.5   ],
               [0.    , 0.    , 0.    , 0.    , 1.    , 0.    , 0.    , 0.    ,
                0.    ]])
        >>> average_difference_matrix(c_elegans)
        array([[0.25  , 0.25  , 0.    , 0.    , 0.    , 0.25  , 0.25  , 0.25  ],
               [0.    , 0.    , 0.5   , 0.5   , 0.    , 0.    , 0.    , 0.    ],
               [0.5   , 0.    , 0.5   , 0.    , 0.5   , 0.    , 0.    , 0.    ],
               [0.    , 0.    , 1.    , 0.    , 0.    , 0.    , 0.    , 0.    ],
               [0.    , 0.3125, 0.3125, 0.3125, 0.3125, 0.3125, 0.    , 0.3125],
               [0.5   , 0.    , 0.    , 0.    , 0.    , 0.5   , 0.5   , 0.    ],
               [1.    , 0.    , 0.    , 0.    , 0.    , 0.    , 0.    , 0.    ],
               [0.    , 0.    , 0.    , 0.    , 0.    , 0.    , 0.5   , 0.5   ]])

    :param states: If None, average over all possible states. Otherwise,
                   providing a list of states will calculate the average over
                   only those states.
    :type states: list or None
    :param calc_trans: Optionally pre-calculate all transitions. Only used
                        when states or weights argument is not None.
    :type calc_trans: bool
    :return: boolean ``numpy`` array
    """
    N = net.size
    Q = np.zeros((N, N))

    if (states is not None) or (weights is not None):
        # explicitly calculate difference matrix for each state

        # optionally pre-calculate transitions
        if calc_trans:
            trans = list(transitions(net))
        else:
            trans = None

        # currently changes state generators to lists.
        # is there a way to avoid this?
        if states is None:
            states = list(net.state_space())
        else:
            states = list(states)

        if weights is None:
            weights = np.ones(len(states))
        else:
            weights = list(weights)

        if np.shape(weights) != (len(states),):
            msg = "Weights must be a 1D array with length same as states"
            raise(ValueError(msg))

        norm = np.sum(weights)
        for i, state in enumerate(states):
            Q += weights[i] * difference_matrix(net, state, trans) / norm

    else:  # make use of sparse connectivity to be more efficient
        state0 = np.zeros(N, dtype=int)

        for i in range(N):
            nodesInfluencingI = list(net.neighbors_in(i))
            for jindex, j in enumerate(nodesInfluencingI):

                # for each state of other nodes, does j matter?
                otherNodes = list(copy.copy(nodesInfluencingI))
                otherNodes.pop(jindex)
                otherNodeStates = _states_limited(otherNodes, state0)
                for state in otherNodeStates:
                    # might be able to do faster by calculating transitions
                    # once for each i also we only need the update for node i

                    # start with two states, one with j on and one with j off
                    jOff = copy.copy(state)
                    jOff[j] = 0
                    jOffNext = net._unsafe_update(jOff)[i]
                    jOn = copy.copy(state)
                    jOn[j] = 1
                    jOnNext = net._unsafe_update(jOn)[i]
                    # are the results different?
                    Q[i, j] += (jOffNext + jOnNext) % 2
                Q[i, j] /= float(len(otherNodeStates))

    return Q


def is_canalizing(net, node_i, neighbor_j):
    """
    Determine whether a given network edge is canalizing: if ``node_i``'s
    value at :math:`t+1` is fully determined when ``neighbor_j``'s value has
    a particular value at :math:`t`, regardless of the values of other nodes,
    then there is a canalizing edge from ``neighbor_j`` to node_i.

    According to (Stauffer 1987), "A rule ... is called forcing, or
    canalizing, if at least one of its :math:`K` arguments has the property
    that the result of the function is already fixed if this argument has
    one particular value, regardless of the values for the :math:`K-1` other
    arguments."  Note that this is a definition for whether a node's rule is
    canalizing, whereas this function calculates whether a specific edge is
    canalizing.  Under this definition, if a node has any incoming canalizing
    edges, then its rule is canalizing.

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> is_canalizing(s_pombe, 1, 2)
        True
        >>> is_canalizing(s_pombe, 2, 1)
        False
        >>> is_canalizing(c_elegans, 7, 7)
        True
        >>> is_canalizing(c_elegans, 1, 3)
        True
        >>> is_canalizing(c_elegans, 4, 3)
        False

    :param net: a :mod:`neet` boolean network
    :param node_i: target node index
    :param neighbor_j: source node index
    :return: ``True`` if the edge ``(neighbor_j, node_i)`` is canalizing, or
             ``None`` if the edge does not exist

    .. seealso:: :func:`canalizing_edges`
    .. seealso:: :func:`canalizing_nodes`
    """
    nodesInfluencingI = list(net.neighbors_in(node_i))

    if (neighbor_j not in nodesInfluencingI) or (node_i not in range(net.size)):
        # can't be canalizing if j has no influence on i
        return None  # or False?
    else:
        jindex = nodesInfluencingI.index(neighbor_j)

        # for every state of other nodes, does j determine i?
        otherNodes = list(copy.copy(nodesInfluencingI))
        otherNodes.pop(jindex)
        otherNodeStates = _states_limited(
            otherNodes, np.zeros(net.size, dtype=int))

        jOnForced, jOffForced = True, True
        jOnForcedValue, jOffForcedValue = None, None
        stateindex = 0
        while (jOnForced or jOffForced) and stateindex < len(otherNodeStates):

            state = otherNodeStates[stateindex]

            # first hold j off
            if jOffForced:
                jOff = copy.copy(state)
                jOff[neighbor_j] = 0
                jOffNext = net.update(jOff)[node_i]
                if jOffForcedValue is None:
                    jOffForcedValue = jOffNext
                elif jOffForcedValue != jOffNext:
                    # then holding j off does not force i
                    jOffForced = False

            # now hold j on
            if jOnForced:
                jOn = copy.copy(state)
                jOn[neighbor_j] = 1
                jOnNext = net.update(jOn)[node_i]
                if jOnForcedValue is None:
                    jOnForcedValue = jOnNext
                elif jOnForcedValue != jOnNext:
                    # then holding j on does not force i
                    jOnForced = False

            stateindex += 1

        # if we have checked all states, then the edge must be forcing
        # print "jOnForced,jOffForced",jOnForced,jOffForced
        return jOnForced or jOffForced


def canalizing_edges(net):
    """
    Return a set of tuples corresponding to the edges in the network that
    are canalizing. Each tuple consists of two node indices, corresponding
    to an edge from the second node to the first node (so that the second node
    controls the first node in a canalizing manner).

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> canalizing_edges(s_pombe)
        {(1, 2), (5, 4), (0, 0), (1, 3), (4, 5), (5, 6), (5, 7), (1, 4), (8, 4), (5, 2), (5, 3)}
        >>> canalizing_edges(c_elegans)
        {(1, 2), (3, 2), (1, 3), (7, 6), (6, 0), (7, 7)}

    :param net: a :mod:`neet` boolean network
    :return: the set of canalizing edges as in the form ``(target, source)``

    .. seealso:: :func:`is_canalizing`
    .. seealso:: :func:`canalizing_nodes`
    """
    canalizingList = []
    for indexi in range(net.size):
        for neighborj in net.neighbors_in(indexi):
            if is_canalizing(net, indexi, neighborj):
                canalizingList.append((indexi, neighborj))
    return set(canalizingList)


def canalizing_nodes(net):
    """
    Find the nodes of the network which have at least one incoming canalizing
    edge.

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> canalizing_nodes(s_pombe)
        {0, 1, 4, 5, 8}
        >>> canalizing_nodes(c_elegans)
        {1, 3, 6, 7}

    :param net: a :mod:`neet` boolean network
    :return: the set indices of nodes with at least one canalizing input edge

    .. seealso:: :func:`is_canalizing`
    .. seealso:: :func:`canalizing_edges`
    """
    nodes = [e[0] for e in canalizing_edges(net)]
    return set(np.unique(nodes))


def lambdaQ(net, **kwargs):
    """
    Calculate sensitivity eigenvalue, the largest eigenvalue of the
    sensitivity matrix :func:`average_difference_matrix`.

    This is analogous to the eigenvalue calculated in [Pomerance2009]_.

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> lambdaQ(s_pombe)
        0.8265021276831896
        >>> lambdaQ(c_elegans)
        1.263099227661824

    :param net: a :mod:`neet` boolean network
    :return: the sensitivity eigenvalue (:math:`\lambda_Q`) of ``net``
    """
    Q = average_difference_matrix(net, **kwargs)
    return max(abs(linalg.eigvals(Q)))


def _fast_encode(state):
    """
    Quickly find encoding of a binary state.

    Same result as ``net.state_space().encode(state)``.
    """
    out = 0
    for bit in state[::-1]:
        out = (out << 1) | bit
    return out


def _boolean_distance(state1, state2):
    """
    Boolean distance between two states.
    """
    out = 0
    for i in range(len(state1)):
        out += state1[i] ^ state2[i]
    return out


def _hamming_neighbors(state):
    """
    Return Hamming neighbors of a boolean state.

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> _hamming_neighbors([0,0,1])
        array([[1, 0, 1],
               [0, 1, 1],
               [0, 0, 0]])
    """
    state = np.asarray(state, dtype=int)
    if len(state.shape) > 1:
        raise(ValueError("state must be 1-dimensional"))
    if not np.array_equal(state % 2, state):
        raise(ValueError("state must be binary"))

    repeat = np.tile(state, (len(state), 1))
    neighbors = (repeat + np.diag(np.ones_like(state))) % 2

    return neighbors


def average_sensitivity(net, states=None, weights=None, calc_trans=True):
    """
    Calculate average Boolean network sensitivity, as defined in
    [Shmulevich2004]_.

    The sensitivity of a Boolean function :math:`f` on state vector :math:`x`
    is the number of Hamming neighbors of :math:`x` on which the function
    value is different than on :math:`x`.

    The average sensitivity is an average taken over initial states.

    .. rubric:: Examples

    .. doctest:: sensitivity

        >>> average_sensitivity(c_elegans)
        1.265625
        >>> average_sensitivity(c_elegans, states=[[0, 0, 0, 0, 0, 0, 0, 0],
        ... [1, 1, 1, 1, 1, 1, 1, 1]])
        ...
        1.5
        >>> average_sensitivity(c_elegans, states=[[0, 0, 0, 0, 0, 0, 0, 0],
        ... [1, 1, 1, 1, 1, 1, 1, 1]],weights=[0.9, 0.1])
        ...
        1.7
        >>> average_sensitivity(c_elegans, states=[[0, 0, 0, 0, 0, 0, 0, 0],
        ... [1, 1, 1, 1, 1, 1, 1, 1]], weights=[9, 1])
        ...
        1.7

    :param net: NEET boolean network
    :param states: Optional list or generator of states. If None, all states
                   are used.
    :param weights: Optional list or generator of weights for each state.
                    If None, each state is equally weighted. If states and
                    weights are both None, an algorithm is used to efficiently
                    make use of sparse connectivity.
    :return: the average sensitivity of ``net``
    """

    if not is_boolean_network(net):
        raise(TypeError("net must be a boolean network"))

    Q = average_difference_matrix(net, states=states, weights=weights,
                                  calc_trans=calc_trans)

    return np.sum(Q) / net.size



"""
Elementary Cellular Automata
============================

The :class:`neet.automata.eca.ECA` class describes an `Elementary Cellular
Automaton <https://en.wikipedia.org/wiki/Elementary_cellular_automaton>`_
with an arbitrary rule. The ``ECA`` class is **not** a fixed sized network.
This means that the size is determined when it is used based on arguments
passed to the relevant methods or functions.

.. rubric:: Examples

.. doctest:: automata

    >>> ca = ECA(30)
    >>> ca.update([0, 0, 1, 0, 0])
    [0, 1, 1, 1, 0]
    >>> ca.update([0, 1, 0])
    [1, 1, 1]
    >>> transitions(ca, size=3)
    [[0, 0, 0], [1, 1, 1], [1, 1, 1], [1, 0, 0], [1, 1, 1], [0, 0, 1], [0, 1, 0], [0, 0, 0]]
"""
import numpy as np
import networkx as nx
from neet.statespace import StateSpace


class ECA(object):
    """
    ECA is a class to represent elementary cellular automaton rules. Each ECA
    contains an 8-bit integral member variable ``code`` representing the
    Wolfram code for the ECA rule and a set of boundary conditions which is
    either ``None``, signifying periodic boundary conditions, or a pair of
    cell states signifying fixed, open boundary conditions.
    """

    def __init__(self, code, boundary=None):
        """
        Construct an elementary cellular automaton rule.

        .. rubric:: Examples

        .. doctest:: automata

            >>> ca = ECA(30)
            >>> ca.code
            30
            >>> ca.boundary
            >>> ca = ECA(30, boundary=(0,0))
            >>> ca.boundary
            (0, 0)

        :param code: the Wolfram code for the ECA
        :type code: int
        :param boundary: the boundary conditions for the CA
        :type boundary: tuple or None
        :raises TypeError: if ``code`` is not an instance of int
        :raises ValueError: if ``code`` is not in :math:`\{0,1,\ldots,255\}`
        :raises TypeError: if ``boundary`` is neither ``None`` or an instance
                           of tuple
        :raises ValueError: if ``boundary`` is a neither ``None`` or a pair of
                            binary states
        """
        self.code = code
        self.boundary = boundary

    @property
    def code(self):
        """
        The Wolfram code of the elementary cellular automaton

        .. rubric:: Examples

        .. doctest:: automata

            >>> eca = ECA(30)
            >>> eca.code
            30
            >>> eca.code = 45
            >>> eca.code
            45
            >>> eca.code = 256
            Traceback (most recent call last):
                ...
            ValueError: invalid ECA code

        :type: int
        :raises TypeError: if ``code`` is not an instance of int
        :raises ValueError: if ``code`` is not in :math:`\{0,1,\ldots,255\}`
        """
        return self.__code

    @code.setter
    def code(self, code):
        if not isinstance(code, int):
            raise TypeError("ECA code is not an int")
        if 255 < code or code < 0:
            raise ValueError("invalid ECA code")
        self.__code = code

    @property
    def boundary(self):
        """
        The boundary conditions of the elemenary cellular automaton

        .. rubric:: Examples

        .. doctest:: automata

            >>> eca = ECA(30)
            >>> eca.boundary
            >>> eca.boundary = (0,1)
            >>> eca.boundary
            (0, 1)
            >>> eca.boundary = None
            >>> eca.boundary
            >>> eca.boundary = [0,1]
            Traceback (most recent call last):
                ...
            TypeError: ECA boundary are neither None nor a tuple

        :type: ``None`` or tuple
        :raises TypeError: if ``boundary`` is neither ``None`` or an instance
                           of tuple
        :raises ValueError: if ``boundary`` is a neither ``None`` or a pair of
                            binary states
        """
        return self.__boundary

    @boundary.setter
    def boundary(self, boundary):
        if boundary and not isinstance(boundary, tuple):
            raise TypeError("ECA boundary are neither None nor a tuple")
        if boundary:
            if len(boundary) != 2:
                raise ValueError("invalid ECA boundary conditions")
            for x in boundary:
                if x != 0 and x != 1:
                    raise ValueError("invalid ECA boundary value")
        self.__boundary = boundary

    def state_space(self, n):
        """
        Return a :class:`neet.statespace.StateSpace` object for a
        lattice of length ``n``.

        .. doctest:: automata

            >>> eca = ECA(30)
            >>> eca.state_space(3)
            <neet.statespace.StateSpace object at 0x...>
            >>> space = eca.state_space(3)
            >>> list(space)
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]

        :param n: the number of nodes in the lattice
        :type n: int
        :returns: :class:`neet.statespace.StateSpace`
        :raises ValueError: if ``n < 1``
        """
        return StateSpace(n, base=2)

    def _unsafe_update(self, lattice, index=None, pin=None, values=None):
        """
        Update the state of the ``lattice``, in place, without
        checking the validity of the arguments.

        .. rubric:: Basic Use:

        .. doctest:: automata

            >>> ca = ECA(30)
            >>> xs = [0,0,1,0,0]
            >>> ca._unsafe_update(xs)
            [0, 1, 1, 1, 0]
            >>> ca.boundary = (1,1)
            >>> ca._unsafe_update([0,0,1,0,0])
            [1, 1, 1, 1, 1]

        .. rubric:: Single-Node Update:

        .. doctest:: automata

            >>> ca.boundary = None
            >>> xs = [0,0,1,0,0]
            >>> ca._unsafe_update(xs, index=1)
            [0, 1, 1, 0, 0]
            >>> xs
            [0, 1, 1, 0, 0]
            >>> ca.boundary = (1,1)
            >>> ca._unsafe_update(xs, index=-1)
            [0, 1, 1, 0, 1]

        .. rubric:: State Pinning:

        .. doctest:: automata

            >>> ca.boundary = None
            >>> xs = [0,0,1,0,0]
            >>> ca._unsafe_update(xs, pin=[-2])
            [0, 1, 1, 0, 0]
            >>> ca.boundary = (1,1)
            >>> ca._unsafe_update(xs, pin=[4])
            [0, 1, 0, 1, 0]

        .. rubric:: Value Fixing:

        .. doctest:: automata

            >>> ca.boundary = None
            >>> xs = [0,0,1,0,0]
            >>> ca._unsafe_update(xs, values={0:1,-2:0})
            [1, 1, 1, 0, 0]
            >>> ca.boundary = (1,1)
            >>> xs = [1,1,1,0,0]
            >>> ca._unsafe_update(xs, values={1:0,-1:0})
            [0, 0, 0, 1, 0]

        :param lattice: the one-dimensional sequence of states
        :type lattice: sequence
        :param index: the index to update (or None)
        :param pin: a sequence of indicies to pin (or None)
        :param values: a dictionary of index-value pairs to fix after update
        :returns: the updated lattice
        """
        pin_states = pin is not None and pin != []
        if self.boundary:
            left = self.__boundary[0]
            right = self.__boundary[1]
        else:
            left = lattice[-1]
            right = lattice[0]
        code = self.code
        if index is None:
            if pin_states:
                pinned = np.asarray(lattice)[pin]
            temp = 2 * left + lattice[0]
            for i in range(1, len(lattice)):
                temp = 7 & (2 * temp + lattice[i])
                lattice[i - 1] = 1 & (code >> temp)
            temp = 7 & (2 * temp + right)
            lattice[-1] = 1 & (code >> temp)
            if pin_states:
                for (j, i) in enumerate(pin):
                    lattice[i] = pinned[j]
        else:
            if index < 0:
                index += len(lattice)

            if index == 0:
                temp = left
            else:
                temp = lattice[index - 1]

            temp = 2 * temp + lattice[index]

            if index + 1 == len(lattice):
                temp = 2 * temp + right
            else:
                temp = 2 * temp + lattice[index + 1]

            lattice[index] = 1 & (code >> (7 & temp))
        if values is not None:
            for key in values:
                lattice[key] = values[key]
        return lattice

    def update(self, lattice, index=None, pin=None, values=None):
        """
        Update the state of the ``lattice`` in place.

        .. rubric:: Basic Use:

        .. doctest:: automata

            >>> ca = ECA(30)
            >>> xs = [0,0,1,0,0]
            >>> ca.update(xs)
            [0, 1, 1, 1, 0]
            >>> ca.boundary = (1,1)
            >>> ca.update([0,0,1,0,0])
            [1, 1, 1, 1, 1]

        .. rubric:: Single-Node Update:

        .. doctest:: automata

            >>> ca.boundary = None
            >>> xs = [0,0,1,0,0]
            >>> ca.update(xs, index=1)
            [0, 1, 1, 0, 0]
            >>> xs
            [0, 1, 1, 0, 0]
            >>> ca.boundary = (1,1)
            >>> ca.update(xs, index=-1)
            [0, 1, 1, 0, 1]

        .. rubric:: State Pinning:

        .. doctest:: automata

            >>> ca.boundary = None
            >>> xs = [0,0,1,0,0]
            >>> ca.update(xs, pin=[-2])
            [0, 1, 1, 0, 0]
            >>> ca.boundary = (1,1)
            >>> ca.update(xs, pin=[4])
            [0, 1, 0, 1, 0]

        .. rubric:: Value Fixing:

        .. doctest:: automata

            >>> ca.boundary = None
            >>> xs = [0,0,1,0,0]
            >>> ca.update(xs, values={0:1,-2:0})
            [1, 1, 1, 0, 0]
            >>> ca.boundary = (1,1)
            >>> xs = [1,1,1,0,0]
            >>> ca.update(xs, values={1:0,-1:0})
            [0, 0, 0, 1, 0]

        .. rubric:: Erroneous Usage:

        .. doctest:: automata

            >>> xs = []
            >>> ca.update(xs)
            Traceback (most recent call last):
            ...
            ValueError: lattice is empty
            >>> xs = [0,0,2,0,0]
            >>> ca.update(xs)
            Traceback (most recent call last):
            ...
            ValueError: invalid value "2" in lattice
            >>> ca.update(xs, index=5)
            Traceback (most recent call last):
            ...
            ValueError: the provided state is not in the ECA's state space
            >>> ca.update([0,0,1,0,0,], index=1, pin=[0])
            Traceback (most recent call last):
            ...
            ValueError: cannot provide both the index and pin arguments
            >>> ca.update([0,0,1,0,0], index=1, values={0:0})
            Traceback (most recent call last):
            ...
            ValueError: cannot provide both the index and values arguments
            >>> ca.update([0,0,1,0,0], pin=[2], values={2:0})
            Traceback (most recent call last):
            ...
            ValueError: cannot set a value for a pinned state
            >>> ca.update([0,0,1,0,0], values={2:2})
            Traceback (most recent call last):
            ...
            ValueError: invalid state in values argument

        :param lattice: the one-dimensional sequence of states
        :param index: the index to update (or None)
        :param pin: a sequence of indicies to pin (or None)
        :param values: a dictionary of index-value pairs to fix after update
        :returns: the updated lattice
        :raises ValueError: if ``lattice`` is not in the ECA's state space
        :raises IndexError: if ``index is not None and index > len(states)``
        :raises ValueError: if ``index`` and ``pin`` are both provided
        :raises ValueError: if ``index`` and ``values`` are both provided
        :raises ValueError: if an element of ``pin`` is a key in ``values``
        :raises ValueError: if a value in ``values`` is not binary (0 or 1)
        """
        size = len(lattice)
        if lattice not in self.state_space(size):
            msg = "the provided state is not in the ECA's state space"
            raise ValueError(msg)

        if index is not None:
            if index < -size:
                raise IndexError("lattice index out of range")
            elif pin is not None and pin != []:
                msg = "cannot provide both the index and pin arguments"
                raise ValueError(msg)
            elif values is not None and values != {}:
                msg = "cannot provide both the index and values arguments"
                raise ValueError(msg)
        elif pin is not None and values is not None:
            for key in values.keys():
                if key in pin:
                    raise ValueError("cannot set a value for a pinned state")
        if values is not None:
            for val in values.values():
                if val != 0 and val != 1:
                    raise ValueError("invalid state in values argument")

        return self._unsafe_update(lattice, index, pin, values)

    def neighbors_in(self, index, size, **kwargs):
        """
        Return the set of all incoming neighbor nodes.

        In the cases of the lattices having fixed boundary conditions, the
        left boundary, being on the left of the leftmost index 0, has an index
        of -1, while the right boundary's index is the size+1. The full state
        of the lattices and the boundaries is equavolent to: `[cell0, cell1,
        ..., cellN, right_boundary, left_boundary]` if it is ever presented as
        a single list in Python.

        :param index: node index
        :param size: size of ECA
        :returns: the set of all node indices which point toward the index node
        :raises ValueError: if `index < 0 or index > n - 1`

        .. rubric:: Basic Use:

        .. doctest:: automata

            >>> net = ECA(30)
            >>> net.neighbors_in(1, size=3)
            {0, 1, 2}
            >>> net.neighbors_in(2, size=3)
            {0, 1, 2}
            >>> net.boundary = (1,1)
            >>> net.neighbors_in(2, size=3)
            {1, 2, 3}
            >>> net.neighbors_in(0, 3)
            {0, 1, -1}

        .. rubric:: Erroneous Usage:

        .. doctest:: automata

            >>> net = ECA(30,boundary=(1, 1))
            >>> net.neighbors_in(5, 3)
            Traceback (most recent call last):
                ...
            ValueError: index must be a non-negative integer less than size
        """
        if not isinstance(size, int):
            raise TypeError("size must be a positive integer")

        if size < 1:
            raise ValueError("size must be a positive integer")

        if not isinstance(index, int):
            raise TypeError("index must be a non-negative integer")

        if index < 0 or index > size - 1:
            msg = "index must be a non-negative integer less than size"
            raise ValueError(msg)

        left, right = index - 1, index + 1

        if left < 0 and self.boundary is None:
            left = size - 1

        if right > size - 1 and self.boundary is None:
            right = 0

        return {left, index, right}

    def neighbors_out(self, index, size):
        """
        Return the set of all outgoing neighbor nodes.

        Fixed boundaries are excluded as they are not affected by internal
        states.

        :param index: node index
        :param size: size of ECA
        :returns: the set of all node indices which point from the index node
        :raises ValueError: if `index < 0 or index > n - 1`

        .. rubric:: Basic Use:

        .. doctest:: automata

            >>> net = ECA(30)
            >>> net.neighbors_out(1, 3)
            {0, 1, 2}
            >>> net.neighbors_out(2, 3)
            {0, 1, 2}
            >>> net.boundary = (1, 1)
            >>> net.neighbors_out(2, 3)
            {1, 2}
            >>> net.neighbors_out(0, 3)
            {0, 1}

        .. rubric:: Erroneous Usage:

        .. doctest:: automata

            >>> net = ECA(30,boundary=(1, 1))
            >>> net.neighbors_out(5, 3)
            Traceback (most recent call last):
                ...
            ValueError: index must be a non-negative integer less than size
        """
        if not isinstance(index, int):
            raise TypeError("index must be a non-negative integer")

        if not isinstance(size, int):
            raise TypeError("size must be a positive integer")

        if size < 1:
            raise ValueError("size must be a positive integer")

        if index < 0 or index > size - 1:
            msg = "index must be a non-negative integer less than size"
            raise ValueError(msg)

        left, right = index - 1, index + 1

        if left < 0:
            left = size - 1 if self.boundary is None else 0

        if right > size - 1:
            right = 0 if self.boundary is None else size - 1

        return {left, index, right}

    def neighbors(self, index, size):
        """
        Return a set of neighbors for a specified node.

        In the cases of the lattices having fixed boundary conditions, the
        left boundary, being on the left of the leftmost index 0, has an index
        of -1, while the right boundary's index is the size+1. The full state
        of the lattices and the boundaries is equavolent to: `[cell0, cell1,
        ..., cellN, right_boundary, left_boundary]` if it is ever presented as
        a single list in Python.

        :param index: node index
        :param size: size of ECA
        :returns: the set of all node indices adjacent to the index node
        :raises ValueError: if `index < 0 or index > n - 1`

        .. rubric:: Basic Use:

        .. doctest:: automata

            >>> net = ECA(30)
            >>> net.neighbors(1, size=3)
            {0, 1, 2}
            >>> net.neighbors(2, size=3)
            {0, 1, 2}
            >>> net.boundary = (1,1)
            >>> net.neighbors(2, size=3)
            {1, 2, 3}
            >>> net.neighbors(0, 3)
            {0, 1, -1}

        .. rubric:: Erroneous Usage:

        .. doctest:: automata

            >>> net = ECA(30,boundary=(1, 1))
            >>> net.neighbors(5, 3)
            Traceback (most recent call last):
                ...
            ValueError: index must be a non-negative integer less than size
        """
        # Outgoing neighbors are a subset of incoming neighbors.
        return self.neighbors_in(index, size)

    def to_networkx_graph(self, size):
        """
        Return networkx graph given neet network. Requires networkx.

        :param size: size of ECA, required if network is an ECA
        :returns: a ``networkx.DiGraph``
        """

        edges = []
        for i in range(size):
            for j in self.neighbors_out(i, size):
                edges.append((i, j))

        return nx.DiGraph(edges, code=self.code, size=size,
                          boundary=self.boundary)

    def draw(self, size, filename=None):
        """
        Output a file with a simple network drawing.

        Requires networkx and pygraphviz.

        Supported image formats are determined by graphviz.  In particular,
        pdf support requires 'cairo' and 'pango' to be installed prior to
        graphviz installation.

        :param filename: filename to write drawing to. Temporary filename will
                         be used if no filename provided.
        :param size: size of ECA, required if network is an ECA
        :returns: a ``pygraphviz`` network drawing
        """
        nx.nx_agraph.view_pygraphviz(
            self.to_networkx_graph(size), prog='circo', path=filename)
